# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import io
import os

import octoprint.plugin
import octoprint.util

from octoprint.events import eventManager, Events

from .destinations import FileDestinations
from .analysis import QueueEntry, AnalysisQueue
from .storage import LocalFileStorage
from .util import AbstractFileWrapper, StreamWrapper, DiskFileWrapper
from octoprint.util import get_fully_qualified_classname as fqcn

from collections import namedtuple

from past.builtins import basestring

ContentTypeMapping = namedtuple("ContentTypeMapping", "extensions, content_type")
ContentTypeDetector = namedtuple("ContentTypeDetector", "extensions, detector")

extensions = dict(
)

def full_extension_tree():
	result = dict(
		# extensions for printable machine code
		machinecode=dict(
			gcode=ContentTypeMapping(["gcode", "gco", "g"], "text/plain")
		)
	)

	def leaf_merger(a, b):
		supported_leaf_types = (ContentTypeMapping, ContentTypeDetector, list)
		if not isinstance(a, supported_leaf_types) or not isinstance(b, supported_leaf_types):
			raise ValueError()

		if isinstance(a, ContentTypeDetector) and isinstance(b, ContentTypeMapping):
			raise ValueError()

		if isinstance(a, ContentTypeMapping) and isinstance(b, ContentTypeDetector):
			raise ValueError()

		a_list = a if isinstance(a, list) else a.extensions
		b_list = b if isinstance(b, list) else b.extensions
		merged = a_list + b_list

		content_type = None
		if isinstance(b, ContentTypeMapping):
			content_type = b.content_type
		elif isinstance(a, ContentTypeMapping):
			content_type = a.content_type

		detector = None
		if isinstance(b, ContentTypeDetector):
			detector = b.detector
		elif isinstance(a, ContentTypeDetector):
			detector = a.detector

		if content_type is not None:
			return ContentTypeMapping(merged, content_type)
		elif detector is not None:
			return ContentTypeDetector(merged, detector)
		else:
			return merged

	slicer_plugins = octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.SlicerPlugin)
	for plugin in slicer_plugins:
		try:
			plugin_result = plugin.get_slicer_extension_tree()
			if plugin_result is None or not isinstance(plugin_result, dict):
				continue
			result = octoprint.util.dict_merge(result, plugin_result, leaf_merger=leaf_merger)
		except Exception:
			logging.getLogger(__name__).exception("Exception while retrieving additional extension "
			                                      "tree entries from SlicerPlugin {name}".format(name=plugin._identifier),
			                                      extra=dict(plugin=plugin._identifier))

	extension_tree_hooks = octoprint.plugin.plugin_manager().get_hooks("octoprint.filemanager.extension_tree")
	for name, hook in extension_tree_hooks.items():
		try:
			hook_result = hook()
			if hook_result is None or not isinstance(hook_result, dict):
				continue
			result = octoprint.util.dict_merge(result, hook_result, leaf_merger=leaf_merger)
		except Exception:
			logging.getLogger(__name__).exception("Exception while retrieving additional extension "
			                                      "tree entries from hook {name}".format(name=name),
			                                      extra=dict(plugin=name))

	return result

def get_extensions(type, subtree=None):
	if subtree is None:
		subtree = full_extension_tree()

	for key, value in subtree.items():
		if key == type:
			return get_all_extensions(subtree=value)
		elif isinstance(value, dict):
			sub_extensions = get_extensions(type, subtree=value)
			if sub_extensions:
				return sub_extensions

	return None

def get_all_extensions(subtree=None):
	if subtree is None:
		subtree = full_extension_tree()

	result = []
	if isinstance(subtree, dict):
		for key, value in subtree.items():
			if isinstance(value, dict):
				result += get_all_extensions(value)
			elif isinstance(value, (ContentTypeMapping, ContentTypeDetector)):
				result += value.extensions
			elif isinstance(value, (list, tuple)):
				result += value
	elif isinstance(subtree, (ContentTypeMapping, ContentTypeDetector)):
		result = subtree.extensions
	elif isinstance(subtree, (list, tuple)):
		result = subtree
	return result

def get_path_for_extension(extension, subtree=None):
	if subtree is None:
		subtree = full_extension_tree()

	for key, value in subtree.items():
		if isinstance(value, (ContentTypeMapping, ContentTypeDetector)) and extension in value.extensions:
			return [key]
		elif isinstance(value, (list, tuple)) and extension in value:
			return [key]
		elif isinstance(value, dict):
			path = get_path_for_extension(extension, subtree=value)
			if path:
				return [key] + path

	return None

def get_content_type_mapping_for_extension(extension, subtree=None):
	if subtree is None:
		subtree = full_extension_tree()

	for key, value in subtree.items():
		content_extension_matches = isinstance(value, (ContentTypeMapping, ContentTypeDetector)) and extension in value.extensions
		list_extension_matches = isinstance(value, (list, tuple)) and extension in value

		if content_extension_matches or list_extension_matches:
			return value
		elif isinstance(value, dict):
			result = get_content_type_mapping_for_extension(extension, subtree=value)
			if result is not None:
				return result

	return None

def valid_extension(extension, type=None):
	if not type:
		return extension in get_all_extensions()
	else:
		extensions = get_extensions(type)
		if extensions:
			return extension in extensions

def valid_file_type(filename, type=None):
	_, extension = os.path.splitext(filename)
	extension = extension[1:].lower()
	return valid_extension(extension, type=type)

def get_file_type(filename):
	_, extension = os.path.splitext(filename)
	extension = extension[1:].lower()
	return get_path_for_extension(extension)

def get_mime_type(filename):
	_, extension = os.path.splitext(filename)
	extension = extension[1:].lower()
	mapping = get_content_type_mapping_for_extension(extension)
	if mapping:
		if isinstance(mapping, ContentTypeMapping) and mapping.content_type is not None:
			return mapping.content_type
		elif isinstance(mapping, ContentTypeDetector) and callable(mapping.detector):
			result = mapping.detector(filename)
			if result is not None:
				return result
	return "application/octet-stream"


class NoSuchStorage(Exception):
	pass


class FileManager(object):
	def __init__(self, analysis_queue, slicing_manager, printer_profile_manager, initial_storage_managers=None):
		self._logger = logging.getLogger(__name__)
		self._analysis_queue = analysis_queue
		self._analysis_queue.register_finish_callback(self._on_analysis_finished)

		self._storage_managers = dict()
		if initial_storage_managers:
			self._storage_managers.update(initial_storage_managers)

		self._slicing_manager = slicing_manager
		self._printer_profile_manager = printer_profile_manager

		import threading
		self._slicing_jobs = dict()
		self._slicing_jobs_mutex = threading.Lock()

		self._slicing_progress_callbacks = []
		self._last_slicing_progress = None

		self._progress_plugins = []
		self._preprocessor_hooks = dict()

		import octoprint.settings
		self._recovery_file = os.path.join(octoprint.settings.settings().getBaseFolder("data"), "print_recovery_data.yaml")
		self._analyzeGcode = octoprint.settings.settings().get(["gcodeAnalysis", "runAt"])

	def initialize(self, process_backlog=False):
		self.reload_plugins()
		if process_backlog:
			self.process_backlog()

	def process_backlog(self):
		# only check for a backlog if gcodeAnalysis is 'idle' or 'always'
		if self._analyzeGcode == "never":
			return

		def worker():
			self._logger.info("Adding backlog items from all storage types to analysis queue...".format(**locals()))
			for storage_type, storage_manager in self._storage_managers.items():
				self._determine_analysis_backlog(storage_type, storage_manager)

		import threading
		thread = threading.Thread(target=worker)
		thread.daemon = True
		thread.start()

	def reload_plugins(self):
		self._progress_plugins = octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.ProgressPlugin)
		self._preprocessor_hooks = octoprint.plugin.plugin_manager().get_hooks("octoprint.filemanager.preprocessor")

	def register_slicingprogress_callback(self, callback):
		self._slicing_progress_callbacks.append(callback)

	def unregister_slicingprogress_callback(self, callback):
		try:
			self._slicing_progress_callbacks.remove(callback)
		except ValueError:
			# callback was not registered
			pass

	def _determine_analysis_backlog(self, storage_type, storage_manager, root=None, high_priority=False):
		counter = 0

		backlog_generator = storage_manager.analysis_backlog
		if root is not None:
			backlog_generator = storage_manager.analysis_backlog_for_path(path=root)

		for entry, path, printer_profile in backlog_generator:
			file_type = get_file_type(path)[-1]
			file_name = storage_manager.split_path(path)

			# we'll use the default printer profile for the backlog since we don't know better
			queue_entry = QueueEntry(file_name, entry, file_type, storage_type, path, self._printer_profile_manager.get_default(), None)
			if self._analysis_queue.enqueue(queue_entry, high_priority=high_priority):
				counter += 1

		if root:
			self._logger.info("Added {counter} items from storage type \"{storage_type}\" and root \"{root}\" to analysis queue".format(**locals()))
		else:
			self._logger.info("Added {counter} items from storage type \"{storage_type}\" to analysis queue".format(**locals()))

	def add_storage(self, storage_type, storage_manager):
		self._storage_managers[storage_type] = storage_manager
		self._determine_analysis_backlog(storage_type, storage_manager)

	def remove_storage(self, type):
		if not type in self._storage_managers:
			return
		del self._storage_managers[type]

	@property
	def registered_storages(self):
		return list(self._storage_managers.keys())

	@property
	def slicing_enabled(self):
		return self._slicing_manager.slicing_enabled

	@property
	def registered_slicers(self):
		return self._slicing_manager.registered_slicers

	@property
	def default_slicer(self):
		return self._slicing_manager.default_slicer

	def analyse(self, destination, path, printer_profile_id=None):
		if not self.file_exists(destination, path):
			return

		if printer_profile_id is None:
			printer_profile = self._printer_profile_manager.get_current_or_default()
		else:
			printer_profile = self._printer_profile_manager.get(printer_profile_id)
			if printer_profile is None:
				printer_profile = self._printer_profile_manager.get_current_or_default()

		queue_entry = self._analysis_queue_entry(destination, path)
		self._analysis_queue.dequeue(queue_entry)

		queue_entry = self._analysis_queue_entry(destination, path, printer_profile=printer_profile)
		if queue_entry:
			return self._analysis_queue.enqueue(queue_entry, high_priority=True)

		return False

	def slice(self, slicer_name, source_location, source_path, dest_location, dest_path,
	          position=None, profile=None, printer_profile_id=None, overrides=None, display=None,
	          callback=None, callback_args=None):
		absolute_source_path = self.path_on_disk(source_location, source_path)

		def stlProcessed(source_location, source_path, tmp_path, dest_location, dest_path, start_time,
		                 printer_profile_id, callback, callback_args, _error=None, _cancelled=False, _analysis=None):
			try:
				if _error:
					eventManager().fire(Events.SLICING_FAILED, dict(slicer=slicer_name,
					                                                stl=source_path,
					                                                stl_location=source_location,
					                                                gcode=dest_path,
					                                                gcode_location=dest_location,
					                                                reason=_error))
				elif _cancelled:
					eventManager().fire(Events.SLICING_CANCELLED, dict(slicer=slicer_name,
					                                                   stl=source_path,
					                                                   stl_location=source_location,
					                                                   gcode=dest_path,
					                                                   gcode_location=dest_location))
				else:
					source_meta = self.get_metadata(source_location, source_path)
					hash = source_meta.get("hash", "n/a")

					import io
					links = [("model", dict(name=source_path))]
					_, stl_name = self.split_path(source_location, source_path)
					file_obj = StreamWrapper(os.path.basename(dest_path),
					                         io.BytesIO(";Generated from {stl_name} (hash: {hash})\n".format(**locals()).encode("ascii", "replace")),
					                         io.FileIO(tmp_path, 'rb'))

					printer_profile = self._printer_profile_manager.get(printer_profile_id)
					self.add_file(dest_location, dest_path, file_obj,
					              display=display, links=links, allow_overwrite=True,
					              printer_profile=printer_profile, analysis=_analysis)

					end_time = octoprint.util.monotonic_time()
					eventManager().fire(Events.SLICING_DONE, dict(slicer=slicer_name,
					                                              stl=source_path,
					                                              stl_location=source_location,
					                                              gcode=dest_path,
					                                              gcode_location=dest_location,
					                                              time=end_time - start_time))

					if callback is not None:
						if callback_args is None:
							callback_args = ()
						callback(*callback_args)
			finally:
				os.remove(tmp_path)

				source_job_key = (source_location, source_path)
				dest_job_key = (dest_location, dest_path)

				with self._slicing_jobs_mutex:
					if source_job_key in self._slicing_jobs:
						del self._slicing_jobs[source_job_key]
					if dest_job_key in self._slicing_jobs:
						del self._slicing_jobs[dest_job_key]

		slicer = self._slicing_manager.get_slicer(slicer_name)

		start_time = octoprint.util.monotonic_time()
		eventManager().fire(Events.SLICING_STARTED, {"slicer": slicer_name,
		                                             "stl": source_path,
		                                             "stl_location": source_location,
		                                             "gcode": dest_path,
		                                             "gcode_location": dest_location,
		                                             "progressAvailable": slicer.get_slicer_properties().get("progress_report", False) if slicer else False})

		import tempfile
		f = tempfile.NamedTemporaryFile(suffix=".gco", delete=False)
		temp_path = f.name
		f.close()

		with self._slicing_jobs_mutex:
			source_job_key = (source_location, source_path)
			dest_job_key = (dest_location, dest_path)
			if dest_job_key in self._slicing_jobs:
				job_slicer_name, job_absolute_source_path, job_temp_path = self._slicing_jobs[dest_job_key]

				self._slicing_manager.cancel_slicing(job_slicer_name, job_absolute_source_path, job_temp_path)
				del self._slicing_jobs[dest_job_key]

			self._slicing_jobs[dest_job_key] = self._slicing_jobs[source_job_key] = (slicer_name, absolute_source_path, temp_path)

		args = (source_location, source_path, temp_path, dest_location, dest_path, start_time, printer_profile_id, callback, callback_args)
		self._slicing_manager.slice(slicer_name,
		                            absolute_source_path,
		                            temp_path,
		                            profile,
		                            stlProcessed,
		                            position=position,
		                            callback_args=args,
		                            overrides=overrides,
		                            printer_profile_id=printer_profile_id,
		                            on_progress=self.on_slicing_progress,
		                            on_progress_args=(slicer_name, source_location, source_path, dest_location, dest_path))

	def on_slicing_progress(self, slicer, source_location, source_path, dest_location, dest_path, _progress=None):
		if not _progress:
			return

		progress_int = int(_progress * 100)
		if self._last_slicing_progress != progress_int:
			self._last_slicing_progress = progress_int
			for callback in self._slicing_progress_callbacks:
				try:
					callback.sendSlicingProgress(slicer, source_location, source_path, dest_location, dest_path, progress_int)
				except Exception:
					self._logger.exception("Exception while pushing slicing progress",
				                           extra=dict(callback=fqcn(callback)))

			if progress_int:
				def call_plugins(slicer, source_location, source_path, dest_location, dest_path, progress):
					for plugin in self._progress_plugins:
						try:
							plugin.on_slicing_progress(slicer, source_location, source_path, dest_location, dest_path, progress)
						except Exception:
							self._logger.exception("Exception while sending slicing progress to plugin %s" % plugin._identifier,
							                       extra=dict(plugin=plugin._identifier))

				import threading
				thread = threading.Thread(target=call_plugins, args=(slicer, source_location, source_path, dest_location, dest_path, progress_int))
				thread.daemon = False
				thread.start()


	def get_busy_files(self):
		return self._slicing_jobs.keys()

	def file_in_path(self, destination, path, file):
		return self._storage(destination).file_in_path(path, file)

	def file_exists(self, destination, path):
		return self._storage(destination).file_exists(path)

	def folder_exists(self, destination, path):
		return self._storage(destination).folder_exists(path)

	def list_files(self, destinations=None, path=None, filter=None, recursive=None):
		if not destinations:
			destinations = list(self._storage_managers.keys())
		if isinstance(destinations, basestring):
			destinations = [destinations]

		result = dict()
		for dst in destinations:
			result[dst] = self._storage_managers[dst].list_files(path=path, filter=filter, recursive=recursive)
		return result

	def add_file(self, destination, path, file_object, links=None, allow_overwrite=False, printer_profile=None, analysis=None, display=None):
		if printer_profile is None:
			printer_profile = self._printer_profile_manager.get_current_or_default()

		for name, hook in self._preprocessor_hooks.items():
			try:
				hook_file_object = hook(path, file_object, links=links, printer_profile=printer_profile, allow_overwrite=allow_overwrite)
			except Exception:
				self._logger.exception("Error when calling preprocessor hook for plugin {}, ignoring".format(name), extra=dict(plugin=name))
				continue

			if hook_file_object is not None:
				file_object = hook_file_object

		queue_entry = self._analysis_queue_entry(destination, path)

		if queue_entry is not None:
			self._analysis_queue.dequeue(queue_entry)

		path_in_storage = self._storage(destination).add_file(path, file_object, links=links, printer_profile=printer_profile, allow_overwrite=allow_overwrite, display=display)

		queue_entry = self._analysis_queue_entry(destination, path_in_storage, printer_profile=printer_profile, analysis=analysis)
		if queue_entry:
			self._analysis_queue.enqueue(queue_entry, high_priority=True)

		_, name = self._storage(destination).split_path(path_in_storage)
		eventManager().fire(Events.FILE_ADDED, dict(storage=destination,
		                                            path=path_in_storage,
		                                            name=name,
		                                            type=get_file_type(name)))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))
		return path_in_storage

	def remove_file(self, destination, path):
		queue_entry = self._analysis_queue_entry(destination, path)
		self._analysis_queue.dequeue(queue_entry)
		self._storage(destination).remove_file(path)

		_, name = self._storage(destination).split_path(path)
		eventManager().fire(Events.FILE_REMOVED, dict(storage=destination,
		                                              path=path,
		                                              name=name,
		                                              type=get_file_type(name)))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))

	def copy_file(self, destination, source, dst):
		path_in_storage = self._storage(destination).copy_file(source, dst)
		if not self.has_analysis(destination, path_in_storage):
			queue_entry = self._analysis_queue_entry(destination, path_in_storage)
			if queue_entry:
				self._analysis_queue.enqueue(queue_entry)

		_, name = self._storage(destination).split_path(path_in_storage)
		eventManager().fire(Events.FILE_ADDED, dict(storage=destination,
		                                            path=path_in_storage,
		                                            name=name,
		                                            type=get_file_type(name)))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))

	def move_file(self, destination, source, dst):
		queue_entry = self._analysis_queue_entry(destination, source)
		self._analysis_queue.dequeue(queue_entry)
		path = self._storage(destination).move_file(source, dst)
		if not self.has_analysis(destination, path):
			queue_entry = self._analysis_queue_entry(destination, path)
			if queue_entry:
				self._analysis_queue.enqueue(queue_entry)

		source_path_in_storage = self._storage(destination).path_in_storage(source)
		_, source_name = self._storage(destination).split_path(source_path_in_storage)
		dst_path_in_storage = self._storage(destination).path_in_storage(dst)
		_, dst_name = self._storage(destination).split_path(dst_path_in_storage)

		eventManager().fire(Events.FILE_REMOVED, dict(storage=destination,
		                                              path=source_path_in_storage,
		                                              name=source_name,
		                                              type=get_file_type(source_name)))
		eventManager().fire(Events.FILE_ADDED, dict(storage=destination,
		                                            path=dst_path_in_storage,
		                                            name=dst_name,
		                                            type=get_file_type(dst_name)))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))

	def add_folder(self, destination, path, ignore_existing=True, display=None):
		path_in_storage = self._storage(destination).add_folder(path, ignore_existing=ignore_existing, display=display)

		_, name = self._storage(destination).split_path(path_in_storage)
		eventManager().fire(Events.FOLDER_ADDED, dict(storage=destination,
		                                              path=path_in_storage,
		                                              name=name))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))
		return path_in_storage

	def remove_folder(self, destination, path, recursive=True):
		self._analysis_queue.dequeue_folder(destination, path)
		self._analysis_queue.pause()
		self._storage(destination).remove_folder(path, recursive=recursive)
		self._analysis_queue.resume()

		_, name = self._storage(destination).split_path(path)
		eventManager().fire(Events.FOLDER_REMOVED, dict(storage=destination,
		                                                path=path,
		                                                name=name))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))

	def copy_folder(self, destination, source, dst):
		path_in_storage = self._storage(destination).copy_folder(source, dst)
		self._determine_analysis_backlog(destination, self._storage(destination), root=path_in_storage)

		_, name = self._storage(destination).split_path(path_in_storage)
		eventManager().fire(Events.FOLDER_ADDED, dict(storage=destination,
		                                              path=path_in_storage,
		                                              name=name))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))

	def move_folder(self, destination, source, dst):
		self._analysis_queue.dequeue_folder(destination, source)
		self._analysis_queue.pause()
		dst_path_in_storage = self._storage(destination).move_folder(source, dst)
		self._determine_analysis_backlog(destination, self._storage(destination), root=dst_path_in_storage)
		self._analysis_queue.resume()

		source_path_in_storage = self._storage(destination).path_in_storage(source)
		_, source_name = self._storage(destination).split_path(source_path_in_storage)
		_, dst_name = self._storage(destination).split_path(dst_path_in_storage)

		eventManager().fire(Events.FOLDER_REMOVED, dict(storage=destination,
		                                                path=source_path_in_storage,
		                                                name=source_name))
		eventManager().fire(Events.FOLDER_ADDED, dict(storage=destination,
		                                              path=dst_path_in_storage,
		                                              name=dst_name))
		eventManager().fire(Events.UPDATED_FILES, dict(type="printables"))

	def has_analysis(self, destination, path):
		return self._storage(destination).has_analysis(path)

	def get_metadata(self, destination, path):
		return self._storage(destination).get_metadata(path)

	def add_link(self, destination, path, rel, data):
		self._storage(destination).add_link(path, rel, data)

	def remove_link(self, destination, path, rel, data):
		self._storage(destination).remove_link(path, rel, data)

	def log_print(self, destination, path, timestamp, print_time, success, printer_profile):
		try:
			if success:
				self._storage(destination).add_history(path, dict(timestamp=timestamp, printTime=print_time, success=success, printerProfile=printer_profile))
			else:
				self._storage(destination).add_history(path, dict(timestamp=timestamp, success=success, printerProfile=printer_profile))
			eventManager().fire(Events.METADATA_STATISTICS_UPDATED, dict(storage=destination, path=path))
		except NoSuchStorage:
			# if there's no storage configured where to log the print, we'll just not log it
			pass

	def save_recovery_data(self, origin, path, pos):
		import time
		import yaml
		from octoprint.util import atomic_write

		data = dict(origin=origin,
		            path=self.path_in_storage(origin, path),
		            pos=pos,
		            date=time.time())
		try:
			with atomic_write(self._recovery_file, mode='wt', max_permissions=0o666) as f:
				yaml.safe_dump(data, stream=f, default_flow_style=False, indent=2, allow_unicode=True)
		except Exception:
			self._logger.exception("Could not write recovery data to file {}".format(self._recovery_file))

	def delete_recovery_data(self):
		if not os.path.isfile(self._recovery_file):
			return

		try:
			os.remove(self._recovery_file)
		except Exception:
			self._logger.exception("Error deleting recovery data file {}".format(self._recovery_file))

	def get_recovery_data(self):
		if not os.path.isfile(self._recovery_file):
			return None

		import yaml
		try:
			with io.open(self._recovery_file, 'rt', encoding='utf-8') as f:
				data = yaml.safe_load(f)
			if not isinstance(data, dict) or not all(map(lambda x: x in data, ("origin", "path", "pos", "date"))):
				raise ValueError("Invalid recovery data structure")
			return data
		except Exception:
			self._logger.exception("Could not read recovery data from file {}".format(self._recovery_file))
			self.delete_recovery_data()

	def set_additional_metadata(self, destination, path, key, data, overwrite=False, merge=False):
		self._storage(destination).set_additional_metadata(path, key, data, overwrite=overwrite, merge=merge)

	def remove_additional_metadata(self, destination, path, key):
		self._storage(destination).remove_additional_metadata(path, key)

	def path_on_disk(self, destination, path):
		return self._storage(destination).path_on_disk(path)

	def canonicalize(self, destination, path):
		return self._storage(destination).canonicalize(path)

	def sanitize(self, destination, path):
		return self._storage(destination).sanitize(path)

	def sanitize_name(self, destination, name):
		return self._storage(destination).sanitize_name(name)

	def sanitize_path(self, destination, path):
		return self._storage(destination).sanitize_path(path)

	def split_path(self, destination, path):
		return self._storage(destination).split_path(path)

	def join_path(self, destination, *path):
		return self._storage(destination).join_path(*path)

	def path_in_storage(self, destination, path):
		return self._storage(destination).path_in_storage(path)

	def last_modified(self, destination, path=None, recursive=False):
		return self._storage(destination).last_modified(path=path, recursive=recursive)

	def _storage(self, destination):
		if not destination in self._storage_managers:
			raise NoSuchStorage("No storage configured for destination {destination}".format(**locals()))
		return self._storage_managers[destination]

	def _add_analysis_result(self, destination, path, result):
		if not destination in self._storage_managers:
			return

		storage_manager = self._storage_managers[destination]
		storage_manager.set_additional_metadata(path, "analysis", result, overwrite=True)

	def _on_analysis_finished(self, entry, result):
		self._add_analysis_result(entry.location, entry.path, result)

	def _analysis_queue_entry(self, destination, path, printer_profile=None, analysis=None):
		if printer_profile is None:
			printer_profile = self._printer_profile_manager.get_current_or_default()

		absolute_path = self._storage(destination).path_on_disk(path)
		_, file_name = self._storage(destination).split_path(path)
		file_type = get_file_type(absolute_path)

		if file_type:
			return QueueEntry(file_name, path, file_type[-1], destination, absolute_path, printer_profile, analysis)
		else:
			return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import click
click.disable_unicode_literals_warning = True

import io
import json

import octoprint_client

from octoprint.cli import get_ctx_obj_option, bulk_options
from octoprint import init_settings, FatalStartupError

from past.builtins import unicode

class JsonStringParamType(click.ParamType):
	name = "json"

	def convert(self, value, param, ctx):
		try:
			return json.loads(value)
		except Exception:
			self.fail("%s is not a valid json string" % value, param, ctx)


def create_client(settings=None, apikey=None, host=None, port=None, httpuser=None, httppass=None, https=False, prefix=None):
	assert(host is not None or settings is not None)
	assert(port is not None or settings is not None)
	assert(apikey is not None or settings is not None)

	if not host:
		host = settings.get(["server", "host"])
		host = host if host != "0.0.0.0" else "127.0.0.1"
	if not port:
		port = settings.getInt(["server", "port"])

	if not apikey:
		apikey = settings.get(["api", "key"])

	baseurl = octoprint_client.build_base_url(https=https,
	                                          httpuser=httpuser,
	                                          httppass=httppass,
	                                          host=host,
	                                          port=port,
	                                          prefix=prefix)

	return octoprint_client.Client(baseurl, apikey)


client_options = bulk_options([
	click.option("--apikey", "-a", type=click.STRING),
	click.option("--host", "-h", type=click.STRING),
	click.option("--port", "-p", type=click.INT),
	click.option("--httpuser", type=click.STRING),
	click.option("--httppass", type=click.STRING),
	click.option("--https", is_flag=True),
	click.option("--prefix", type=click.STRING)
])
"""Common options to configure an API client."""


@click.group()
def client_commands():
	pass


@client_commands.group("client", context_settings=dict(ignore_unknown_options=True))
@client_options
@click.pass_context
def client(ctx, apikey, host, port, httpuser, httppass, https, prefix):
	"""Basic API client."""
	try:
		settings = None
		if not host or not port or not apikey:
			settings = init_settings(get_ctx_obj_option(ctx, "basedir", None), get_ctx_obj_option(ctx, "configfile", None))

		ctx.obj.client = create_client(settings=settings,
		                               apikey=apikey,
		                               host=host,
		                               port=port,
		                               httpuser=httpuser,
		                               httppass=httppass,
		                               https=https,
		                               prefix=prefix)

	except FatalStartupError as e:
		click.echo(e.message, err=True)
		click.echo("There was a fatal error initializing the client.", err=True)
		ctx.exit(-1)


def log_response(response, status_code=True, body=True, headers=False):
	if status_code:
		click.echo("Status Code: {}".format(response.status_code))
	if headers:
		for header, value in response.headers.items():
			click.echo("{}: {}".format(header, value))
		click.echo()
	if body:
		click.echo(response.text)


@client.command("get")
@click.argument("path")
@click.option("--timeout", type=float, default=None)
@click.pass_context
def get(ctx, path, timeout):
	"""Performs a GET request against the specified server path."""
	r = ctx.obj.client.get(path, timeout=timeout)
	log_response(r)


@client.command("post_json")
@click.argument("path")
@click.argument("data", type=JsonStringParamType())
@click.option("--timeout", type=float, default=None)
@click.pass_context
def post_json(ctx, path, data, timeout):
	"""POSTs JSON data to the specified server path."""
	r = ctx.obj.client.post_json(path, data, timeout=timeout)
	log_response(r)


@client.command("patch_json")
@click.argument("path")
@click.argument("data", type=JsonStringParamType())
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def patch_json(ctx, path, data, timeout):
	"""PATCHes JSON data to the specified server path."""
	r = ctx.obj.client.patch(path, data, encoding="json", timeout=timeout)
	log_response(r)


@client.command("post_from_file")
@click.argument("path")
@click.argument("file_path", type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@click.option("--json", is_flag=True)
@click.option("--yaml", is_flag=True)
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def post_from_file(ctx, path, file_path, json_flag, yaml_flag, timeout):
	"""POSTs JSON data to the specified server path, taking the data from the specified file."""
	if json_flag or yaml_flag:
		if json_flag:
			with io.open(file_path, 'rt') as fp:
				data = json.load(fp)
		else:
			import yaml
			with io.open(file_path, 'rt') as fp:
				data = yaml.safe_load(fp)

		r = ctx.obj.client.post_json(path, data, timeout=timeout)
	else:
		with io.open(file_path, 'rb') as fp:
			data = fp.read()

		r = ctx.obj.client.post(path, data, timeout=timeout)

	log_response(r)


@client.command("command")
@click.argument("path")
@click.argument("command")
@click.option("--str", "-s", "str_params", multiple=True, nargs=2, type=click.Tuple([unicode, unicode]))
@click.option("--int", "-i", "int_params", multiple=True, nargs=2, type=click.Tuple([unicode, int]))
@click.option("--float", "-f", "float_params", multiple=True, nargs=2, type=click.Tuple([unicode, float]))
@click.option("--bool", "-b", "bool_params", multiple=True, nargs=2, type=click.Tuple([unicode, bool]))
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def command(ctx, path, command, str_params, int_params, float_params, bool_params, timeout):
	"""Sends a JSON command to the specified server path."""
	data = dict()
	params = str_params + int_params + float_params + bool_params
	for param in params:
		data[param[0]] = param[1]
	r = ctx.obj.client.post_command(path, command, additional=data, timeout=timeout)
	log_response(r, body=False)


@client.command("upload")
@click.argument("path")
@click.argument("file_path", type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@click.option("--parameter", "-P", "params", multiple=True, nargs=2, type=click.Tuple([unicode, unicode]))
@click.option("--file-name", type=click.STRING)
@click.option("--content-type", type=click.STRING)
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def upload(ctx, path, file_path, params, file_name, content_type, timeout):
	"""Uploads the specified file to the specified server path."""
	data = dict()
	for param in params:
		data[param[0]] = param[1]

	r = ctx.obj.client.upload(path, file_path,
	                          additional=data, file_name=file_name, content_type=content_type, timeout=timeout)
	log_response(r)


@client.command("delete")
@click.argument("path")
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def delete(ctx, path, timeout):
	"""Sends a DELETE request to the specified server path."""
	r = ctx.obj.client.delete(path, timeout=timeout)
	log_response(r)


@client.command("listen")
@click.pass_context
def listen(ctx):
	def on_connect(ws):
		click.echo(">>> Connected!")

	def on_close(ws):
		click.echo(">>> Connection closed!")

	def on_error(ws, error):
		click.echo("!!! Error: {}".format(error))

	def on_heartbeat(ws):
		click.echo("<3")

	def on_message(ws, message_type, message_payload):
		click.echo("Message: {}, Payload: {}".format(message_type, json.dumps(message_payload)))

	socket = ctx.obj.client.create_socket(on_connect=on_connect,
	                                      on_close=on_close,
	                                      on_error=on_error,
	                                      on_heartbeat=on_heartbeat,
	                                      on_message=on_message)

	click.echo(">>> Waiting for client to exit")
	try:
		socket.wait()
	finally:
		click.echo(">>> Goodbye...")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
Generic linux daemon base class

Originally from http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/#c35
"""

import io, sys, os, time, signal

class Daemon:
	"""
	A generic daemon class.

	Usage: subclass the daemon class and override the run() method.

	If you want to log the output to someplace different that stdout and stderr,
	also override the echo() and error() methods.
	"""

	def __init__(self, pidfile):
		self.pidfile = pidfile

	def _daemonize(self):
		"""Daemonize class. UNIX double fork mechanism."""

		self._double_fork()
		self._redirect_io()

		# write pidfile
		pid = str(os.getpid())
		self.set_pid(pid)

	def _double_fork(self):
		try:
			pid = os.fork()
			if pid > 0:
				# exit first parent
				sys.exit(0)
		except OSError as err:
			self.error("First fork failed: {}".format(str(err)))
			sys.exit(1)

		# decouple from parent environment
		os.chdir('/')
		os.setsid()
		os.umask(0o002)

		# do second fork
		try:
			pid = os.fork()
			if pid > 0:
				# exit from second parent
				sys.exit(0)
		except OSError as err:
			self.error("Second fork failed: {}".format(str(err)))
			sys.exit(1)

	def _redirect_io(self):
		# redirect standard file descriptors
		sys.stdout.flush()
		sys.stderr.flush()
		si = io.open(os.devnull, 'rt', encoding='utf-8')
		so = io.open(os.devnull, 'at+', encoding='utf-8')
		se = io.open(os.devnull, 'at+', encoding='utf-8')

		os.dup2(si.fileno(), sys.stdin.fileno())
		os.dup2(so.fileno(), sys.stdout.fileno())
		os.dup2(se.fileno(), sys.stderr.fileno())

	def terminated(self):
		self.remove_pidfile()

	def start(self):
		"""Start the daemon."""

		# Check for a pidfile to see if the daemon already runs
		pid = self.get_pid()
		if pid:
			self.error("pidfile {} already exist. Is the daemon already running?".format(self.pidfile))
			sys.exit(1)

		self.echo("Starting daemon...")

		# Start the daemon
		self._daemonize()
		self.run()

	def stop(self, check_running=True):
		"""Stop the daemon."""
		pid = self.get_pid()
		if not pid:
			if not check_running:
				return
			self.error("pidfile {} does not exist. Is the daemon really running?".format(self.pidfile))
			sys.exit(1)

		self.echo("Stopping daemon...")

		# Try killing the daemon process
		try:
			while 1:
				os.kill(pid, signal.SIGTERM)
				time.sleep(0.1)
		except OSError as err:
			e = str(err.args)
			if e.find("No such process") > 0:
				self.remove_pidfile()
			else:
				self.error(e)
				sys.exit(1)

	def restart(self):
		"""Restart the daemon."""
		self.stop(check_running=False)
		self.start()

	def status(self):
		"""Prints the daemon status."""
		if self.is_running():
			self.echo("Daemon is running")
		else:
			self.echo("Daemon is not running")

	def is_running(self):
		"""Check if a process is running under the specified pid."""
		pid = self.get_pid()
		if pid is None:
			return False

		try:
			os.kill(pid, 0)
		except OSError:
			try:
				self.remove_pidfile()
			except Exception:
				self.error("Daemon found not running, but could not remove stale pidfile")
			return False
		else:
			return True

	def get_pid(self):
		"""Get the pid from the pidfile."""
		try:
			with io.open(self.pidfile,'rt', encoding='utf-8') as pf:
				pid = int(pf.read().strip())
		except (IOError, ValueError):
			pid = None
		return pid

	def set_pid(self, pid):
		"""Write the pid to the pidfile."""
		with io.open(self.pidfile,'wt+', encoding='utf-8') as f:
			f.write(str(pid) + '\n')

	def remove_pidfile(self):
		"""Removes the pidfile."""
		if os.path.isfile(self.pidfile):
			os.remove(self.pidfile)

	def run(self):
		"""You should override this method when you subclass Daemon.

		It will be called after the process has been daemonized by
		start() or restart()."""

		raise NotImplementedError()

	@classmethod
	def echo(cls, line):
		print(line)

	@classmethod
	def error(cls, line):
		print(line, file=sys.stderr)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging

from flask import request, jsonify, make_response, abort
from flask_login import current_user
from werkzeug.exceptions import BadRequest

from octoprint.events import eventManager, Events
from octoprint.settings import settings, valid_boolean_trues

from octoprint.server import admin_permission, printer, pluginManager
from octoprint.server.api import api, NO_CONTENT
from octoprint.server.util.flask import no_firstrun_access, with_revalidation_checking
from octoprint.access.permissions import Permissions

import octoprint.plugin
import octoprint.util

#~~ settings

def _lastmodified():
	return settings().last_modified

def _etag(lm=None):
	if lm is None:
		lm = _lastmodified()

	connection_options = printer.__class__.get_connection_options()
	plugins = sorted(octoprint.plugin.plugin_manager().enabled_plugins)
	plugin_settings = _get_plugin_settings()

	from collections import OrderedDict
	sorted_plugin_settings = OrderedDict()
	for key in sorted(plugin_settings.keys()):
		sorted_plugin_settings[key] = plugin_settings.get(key, dict())

	if current_user is not None and not current_user.is_anonymous:
		roles = sorted(current_user.permissions)
	else:
		roles = []

	import hashlib
	hash = hashlib.sha1()
	def hash_update(value):
		value = value.encode('utf-8')
		hash.update(value)

	# last modified timestamp
	hash_update(str(lm))

	# effective config from config.yaml + overlays
	hash_update(repr(settings().effective))

	# might duplicate settings().effective, but plugins might also inject additional keys into the settings
	# output that are not stored in config.yaml
	hash_update(repr(sorted_plugin_settings))

	# connection options are also part of the settings
	hash_update(repr(connection_options))

	# if the list of plugins changes, the settings structure changes too
	hash_update(repr(plugins))

	# and likewise if the role of the user changes
	hash_update(repr(roles))

	return hash.hexdigest()

@api.route("/settings", methods=["GET"])
@with_revalidation_checking(etag_factory=_etag,
                            lastmodified_factory=_lastmodified,
                            unless=lambda: request.values.get("force", "false") in valid_boolean_trues or settings().getBoolean(["server", "firstRun"]))
def getSettings():
	if not Permissions.SETTINGS_READ.can() and not settings().getBoolean(["server", "firstRun"]):
		abort(403)

	s = settings()

	connectionOptions = printer.__class__.get_connection_options()

	# NOTE: Remember to adjust the docs of the data model on the Settings API if anything
	# is changed, added or removed here

	data = {
		"api": {
			"key": s.get(["api", "key"]) if Permissions.ADMIN.can() else None,
			"allowCrossOrigin": s.get(["api", "allowCrossOrigin"])
		},
		"appearance": {
			"name": s.get(["appearance", "name"]),
			"color": s.get(["appearance", "color"]),
			"colorTransparent": s.getBoolean(["appearance", "colorTransparent"]),
			"colorIcon": s.getBoolean(["appearance", "colorIcon"]),
			"defaultLanguage": s.get(["appearance", "defaultLanguage"]),
			"showFahrenheitAlso": s.getBoolean(["appearance", "showFahrenheitAlso"]),
			"fuzzyTimes": s.getBoolean(["appearance", "fuzzyTimes"]),
			"closeModalsWithClick": s.getBoolean(["appearance", "closeModalsWithClick"])
		},
		"printer": {
			"defaultExtrusionLength": s.getInt(["printerParameters", "defaultExtrusionLength"])
		},
		"webcam": {
			"webcamEnabled": s.getBoolean(["webcam", "webcamEnabled"]),
			"timelapseEnabled": s.getBoolean(["webcam", "timelapseEnabled"]),
			"streamUrl": s.get(["webcam", "stream"]),
			"streamRatio": s.get(["webcam", "streamRatio"]),
			"streamTimeout": s.getInt(["webcam", "streamTimeout"]),
			"snapshotUrl": s.get(["webcam", "snapshot"]),
			"snapshotTimeout": s.getInt(["webcam", "snapshotTimeout"]),
			"snapshotSslValidation": s.getBoolean(["webcam", "snapshotSslValidation"]),
			"ffmpegPath": s.get(["webcam", "ffmpeg"]),
			"bitrate": s.get(["webcam", "bitrate"]),
			"ffmpegThreads": s.get(["webcam", "ffmpegThreads"]),
			"ffmpegVideoCodec": s.get(["webcam", "ffmpegVideoCodec"]),
			"watermark": s.getBoolean(["webcam", "watermark"]),
			"flipH": s.getBoolean(["webcam", "flipH"]),
			"flipV": s.getBoolean(["webcam", "flipV"]),
			"rotate90": s.getBoolean(["webcam", "rotate90"])
		},
		"feature": {
			"gcodeViewer": s.getBoolean(["gcodeViewer", "enabled"]),
			"sizeThreshold": s.getInt(["gcodeViewer", "sizeThreshold"]),
			"mobileSizeThreshold": s.getInt(["gcodeViewer", "mobileSizeThreshold"]),
			"temperatureGraph": s.getBoolean(["feature", "temperatureGraph"]),
			"sdSupport": s.getBoolean(["feature", "sdSupport"]),
			"keyboardControl": s.getBoolean(["feature", "keyboardControl"]),
			"pollWatched": s.getBoolean(["feature", "pollWatched"]),
			"modelSizeDetection": s.getBoolean(["feature", "modelSizeDetection"]),
			"printStartConfirmation": s.getBoolean(["feature", "printStartConfirmation"]),
			"printCancelConfirmation": s.getBoolean(["feature", "printCancelConfirmation"]),
			"g90InfluencesExtruder": s.getBoolean(["feature", "g90InfluencesExtruder"]),
			"autoUppercaseBlacklist": s.get(["feature", "autoUppercaseBlacklist"])
		},
		"gcodeAnalysis": {
			"runAt": s.get(["gcodeAnalysis", "runAt"])
		},
		"serial": {
			"port": connectionOptions["portPreference"],
			"baudrate": connectionOptions["baudratePreference"],
			"exclusive": s.getBoolean(["serial", "exclusive"]),
			"portOptions": connectionOptions["ports"],
			"baudrateOptions": connectionOptions["baudrates"],
			"autoconnect": s.getBoolean(["serial", "autoconnect"]),
			"timeoutConnection": s.getFloat(["serial", "timeout", "connection"]),
			"timeoutDetection": s.getFloat(["serial", "timeout", "detection"]),
			"timeoutCommunication": s.getFloat(["serial", "timeout", "communication"]),
			"timeoutCommunicationBusy": s.getFloat(["serial", "timeout", "communicationBusy"]),
			"timeoutTemperature": s.getFloat(["serial", "timeout", "temperature"]),
			"timeoutTemperatureTargetSet": s.getFloat(["serial", "timeout", "temperatureTargetSet"]),
			"timeoutTemperatureAutoreport": s.getFloat(["serial", "timeout", "temperatureAutoreport"]),
			"timeoutSdStatus": s.getFloat(["serial", "timeout", "sdStatus"]),
			"timeoutSdStatusAutoreport": s.getFloat(["serial", "timeout", "sdStatusAutoreport"]),
			"timeoutBaudrateDetectionPause": s.getFloat(["serial", "timeout", "baudrateDetectionPause"]),
			"timeoutPositionLogWait": s.getFloat(["serial", "timeout", "positionLogWait"]),
			"log": s.getBoolean(["serial", "log"]),
			"additionalPorts": s.get(["serial", "additionalPorts"]),
			"additionalBaudrates": s.get(["serial", "additionalBaudrates"]),
			"longRunningCommands": s.get(["serial", "longRunningCommands"]),
			"checksumRequiringCommands": s.get(["serial", "checksumRequiringCommands"]),
			"blockedCommands": s.get(["serial", "blockedCommands"]),
			"pausingCommands": s.get(["serial", "pausingCommands"]),
			"emergencyCommands": s.get(["serial", "emergencyCommands"]),
			"helloCommand": s.get(["serial", "helloCommand"]),
			"ignoreErrorsFromFirmware": s.getBoolean(["serial", "ignoreErrorsFromFirmware"]),
			"disconnectOnErrors": s.getBoolean(["serial", "disconnectOnErrors"]),
			"triggerOkForM29": s.getBoolean(["serial", "triggerOkForM29"]),
			"logPositionOnPause": s.getBoolean(["serial", "logPositionOnPause"]),
			"logPositionOnCancel": s.getBoolean(["serial", "logPositionOnCancel"]),
			"abortHeatupOnCancel": s.getBoolean(["serial", "abortHeatupOnCancel"]),
			"supportResendsWithoutOk": s.get(["serial", "supportResendsWithoutOk"]),
			"waitForStart": s.getBoolean(["serial", "waitForStartOnConnect"]),
			"alwaysSendChecksum": s.getBoolean(["serial", "alwaysSendChecksum"]),
			"neverSendChecksum": s.getBoolean(["serial", "neverSendChecksum"]),
			"sdRelativePath": s.getBoolean(["serial", "sdRelativePath"]),
			"sdAlwaysAvailable": s.getBoolean(["serial", "sdAlwaysAvailable"]),
			"swallowOkAfterResend": s.getBoolean(["serial", "swallowOkAfterResend"]),
			"repetierTargetTemp": s.getBoolean(["serial", "repetierTargetTemp"]),
			"externalHeatupDetection": s.getBoolean(["serial", "externalHeatupDetection"]),
			"ignoreIdenticalResends": s.getBoolean(["serial", "ignoreIdenticalResends"]),
			"firmwareDetection": s.getBoolean(["serial", "firmwareDetection"]),
			"blockWhileDwelling": s.getBoolean(["serial", "blockWhileDwelling"]),
			"useParityWorkaround": s.get(["serial", "useParityWorkaround"]),
			"sendM112OnError": s.getBoolean(["serial", "sendM112OnError"]),
			"ackMax": s.getInt(["serial", "ackMax"]),
			"maxTimeoutsIdle": s.getInt(["serial", "maxCommunicationTimeouts", "idle"]),
			"maxTimeoutsPrinting": s.getInt(["serial", "maxCommunicationTimeouts", "printing"]),
			"maxTimeoutsLong": s.getInt(["serial", "maxCommunicationTimeouts", "long"]),
			"capAutoreportTemp": s.getBoolean(["serial", "capabilities", "autoreport_temp"]),
			"capAutoreportSdStatus": s.getBoolean(["serial", "capabilities", "autoreport_sdstatus"]),
			"capBusyProtocol": s.getBoolean(["serial", "capabilities", "busy_protocol"]),
			"capEmergencyParser": s.getBoolean(["serial", "capabilities", "emergency_parser"])
		},
		"folder": {
			"uploads": s.getBaseFolder("uploads"),
			"timelapse": s.getBaseFolder("timelapse"),
			"timelapseTmp": s.getBaseFolder("timelapse_tmp"),
			"logs": s.getBaseFolder("logs"),
			"watched": s.getBaseFolder("watched")
		},
		"temperature": {
			"profiles": s.get(["temperature", "profiles"]),
			"cutoff": s.getInt(["temperature", "cutoff"]),
			"sendAutomatically": s.getBoolean(["temperature", "sendAutomatically"]),
			"sendAutomaticallyAfter": s.getInt(["temperature", "sendAutomaticallyAfter"], min=0, max=30),
		},
		"system": {
			"actions": s.get(["system", "actions"]),
			"events": s.get(["system", "events"])
		},
		"terminalFilters": s.get(["terminalFilters"]),
		"scripts": {
			"gcode": {
				"afterPrinterConnected": None,
				"beforePrinterDisconnected": None,
				"beforePrintStarted": None,
				"afterPrintCancelled": None,
				"afterPrintDone": None,
				"beforePrintPaused": None,
				"afterPrintResumed": None,
				"beforeToolChange": None,
				"afterToolChange": None,
				"snippets": dict()
			}
		},
		"server": {
			"commands": {
				"systemShutdownCommand": s.get(["server", "commands", "systemShutdownCommand"]),
				"systemRestartCommand": s.get(["server", "commands", "systemRestartCommand"]),
				"serverRestartCommand": s.get(["server", "commands", "serverRestartCommand"])
			},
			"diskspace": {
				"warning": s.getInt(["server", "diskspace", "warning"]),
				"critical": s.getInt(["server", "diskspace", "critical"])
			},
			"onlineCheck": {
				"enabled": s.getBoolean(["server", "onlineCheck", "enabled"]),
				"interval": int(s.getInt(["server", "onlineCheck", "interval"]) / 60),
				"host": s.get(["server", "onlineCheck", "host"]),
				"port": s.getInt(["server", "onlineCheck", "port"])
			},
			"pluginBlacklist": {
				"enabled": s.getBoolean(["server", "pluginBlacklist", "enabled"]),
				"url": s.get(["server", "pluginBlacklist", "url"]),
				"ttl": int(s.getInt(["server", "pluginBlacklist", "ttl"]) / 60)
			},
			"allowFraming": s.getBoolean(["server", "allowFraming"])
		}
	}

	gcode_scripts = s.listScripts("gcode")
	if gcode_scripts:
		data["scripts"] = dict(gcode=dict())
		for name in gcode_scripts:
			data["scripts"]["gcode"][name] = s.loadScript("gcode", name, source=True)

	plugin_settings = _get_plugin_settings()
	if len(plugin_settings):
		data["plugins"] = plugin_settings

	return jsonify(data)


def _get_plugin_settings():
	logger = logging.getLogger(__name__)

	data = dict()

	def process_plugin_result(name, result):
		if result:
			try:
				jsonify(test=result)
			except Exception:
				logger.exception("Error while jsonifying settings from plugin {}, please contact the plugin author about this".format(name))
				raise
			else:
				if "__enabled" in result:
					del result["__enabled"]
				data[name] = result

	for plugin in octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.SettingsPlugin):
		try:
			result = plugin.on_settings_load()
			process_plugin_result(plugin._identifier, result)
		except Exception:
			logger.exception("Could not load settings for plugin {name} ({version})".format(version=plugin._plugin_version,
			                                                                                name=plugin._plugin_name),
			                 extra=dict(plugin=plugin._identifier))

	return data


@api.route("/settings", methods=["POST"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def setSettings():
	if not "application/json" in request.headers["Content-Type"]:
		return make_response("Expected content-type JSON", 400)

	try:
		data = request.get_json()
	except BadRequest:
		return make_response("Malformed JSON body in request", 400)

	if data is None:
		return make_response("Malformed JSON body in request", 400)

	if not isinstance(data, dict):
		return make_response("Malformed request, need settings dictionary, "
		                     "got a {} instead: {!r}".format(type(data).__name__, data), 400)

	response = _saveSettings(data)
	if response:
		return response
	return getSettings()


@api.route("/settings/apikey", methods=["POST"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def generateApiKey():
	apikey = settings().generateApiKey()
	return jsonify(apikey=apikey)


@api.route("/settings/apikey", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def deleteApiKey():
	settings().deleteApiKey()
	return NO_CONTENT

@api.route("/settings/templates", methods=["GET"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def fetchTemplateData():
	from octoprint.server.views import fetch_template_data

	refresh = request.values.get("refresh", "false") in valid_boolean_trues
	templates, _, _ = fetch_template_data(refresh=refresh)

	result = dict()
	for tt in templates:
		result[tt] = []
		for key in templates[tt]["order"]:
			entry = templates[tt]["entries"].get(key)
			if not entry:
				continue

			if isinstance(entry, dict):
				name = key
			else:
				name, entry = entry

			data = dict(id=key, name=name)

			if entry and "_plugin" in entry:
				plugin = pluginManager.get_plugin_info(entry["_plugin"], require_enabled=False)
				data["plugin_id"] = plugin.key
				data["plugin_name"] = plugin.name

			result[tt].append(data)

	return jsonify(order=result)

def _saveSettings(data):
	logger = logging.getLogger(__name__)

	s = settings()

	# NOTE: Remember to adjust the docs of the data model on the Settings API if anything
	# is changed, added or removed here

	if "folder" in data:
		try:
			if "uploads" in data["folder"]: s.setBaseFolder("uploads", data["folder"]["uploads"])
			if "timelapse" in data["folder"]: s.setBaseFolder("timelapse", data["folder"]["timelapse"])
			if "timelapseTmp" in data["folder"]: s.setBaseFolder("timelapse_tmp", data["folder"]["timelapseTmp"])
			if "logs" in data["folder"]: s.setBaseFolder("logs", data["folder"]["logs"])
			if "watched" in data["folder"]: s.setBaseFolder("watched", data["folder"]["watched"])
		except IOError:
			return make_response("One of the configured folders is invalid", 400)

	if "api" in data:
		if "allowCrossOrigin" in data["api"]: s.setBoolean(["api", "allowCrossOrigin"], data["api"]["allowCrossOrigin"])

	if "appearance" in data:
		if "name" in data["appearance"]: s.set(["appearance", "name"], data["appearance"]["name"])
		if "color" in data["appearance"]: s.set(["appearance", "color"], data["appearance"]["color"])
		if "colorTransparent" in data["appearance"]: s.setBoolean(["appearance", "colorTransparent"], data["appearance"]["colorTransparent"])
		if "colorIcon" in data["appearance"]: s.setBoolean(["appearance", "colorIcon"], data["appearance"]["colorIcon"])
		if "defaultLanguage" in data["appearance"]: s.set(["appearance", "defaultLanguage"], data["appearance"]["defaultLanguage"])
		if "showFahrenheitAlso" in data["appearance"]: s.setBoolean(["appearance", "showFahrenheitAlso"], data["appearance"]["showFahrenheitAlso"])
		if "fuzzyTimes" in data["appearance"]: s.setBoolean(["appearance", "fuzzyTimes"], data["appearance"]["fuzzyTimes"])
		if "closeModalsWithClick" in data["appearance"]: s.setBoolean(["appearance", "closeModalsWithClick"], data["appearance"]["closeModalsWithClick"])

	if "printer" in data:
		if "defaultExtrusionLength" in data["printer"]: s.setInt(["printerParameters", "defaultExtrusionLength"], data["printer"]["defaultExtrusionLength"])

	if "webcam" in data:
		if "webcamEnabled" in data["webcam"]: s.setBoolean(["webcam", "webcamEnabled"], data["webcam"]["webcamEnabled"])
		if "timelapseEnabled" in data["webcam"]: s.setBoolean(["webcam", "timelapseEnabled"], data["webcam"]["timelapseEnabled"])
		if "streamUrl" in data["webcam"]: s.set(["webcam", "stream"], data["webcam"]["streamUrl"])
		if "streamRatio" in data["webcam"] and data["webcam"]["streamRatio"] in ("16:9", "4:3"): s.set(["webcam", "streamRatio"], data["webcam"]["streamRatio"])
		if "streamTimeout" in data["webcam"]: s.setInt(["webcam", "streamTimeout"], data["webcam"]["streamTimeout"])
		if "snapshotUrl" in data["webcam"]: s.set(["webcam", "snapshot"], data["webcam"]["snapshotUrl"])
		if "snapshotTimeout" in data["webcam"]: s.setInt(["webcam", "snapshotTimeout"], data["webcam"]["snapshotTimeout"])
		if "snapshotSslValidation" in data["webcam"]: s.setBoolean(["webcam", "snapshotSslValidation"], data["webcam"]["snapshotSslValidation"])
		if "ffmpegPath" in data["webcam"]: s.set(["webcam", "ffmpeg"], data["webcam"]["ffmpegPath"])
		if "bitrate" in data["webcam"]: s.set(["webcam", "bitrate"], data["webcam"]["bitrate"])
		if "ffmpegThreads" in data["webcam"]: s.setInt(["webcam", "ffmpegThreads"], data["webcam"]["ffmpegThreads"])
		if "ffmpegVideoCodec" in data["webcam"] and data["webcam"]["ffmpegVideoCodec"] in ("mpeg2video", "libx264"): s.set(["webcam", "ffmpegVideoCodec"], data["webcam"]["ffmpegVideoCodec"])
		if "watermark" in data["webcam"]: s.setBoolean(["webcam", "watermark"], data["webcam"]["watermark"])
		if "flipH" in data["webcam"]: s.setBoolean(["webcam", "flipH"], data["webcam"]["flipH"])
		if "flipV" in data["webcam"]: s.setBoolean(["webcam", "flipV"], data["webcam"]["flipV"])
		if "rotate90" in data["webcam"]: s.setBoolean(["webcam", "rotate90"], data["webcam"]["rotate90"])

	if "feature" in data:
		if "gcodeViewer" in data["feature"]: s.setBoolean(["gcodeViewer", "enabled"], data["feature"]["gcodeViewer"])
		if "sizeThreshold" in data["feature"]: s.setInt(["gcodeViewer", "sizeThreshold"], data["feature"]["sizeThreshold"])
		if "mobileSizeThreshold" in data["feature"]: s.setInt(["gcodeViewer", "mobileSizeThreshold"], data["feature"]["mobileSizeThreshold"])
		if "temperatureGraph" in data["feature"]: s.setBoolean(["feature", "temperatureGraph"], data["feature"]["temperatureGraph"])
		if "sdSupport" in data["feature"]: s.setBoolean(["feature", "sdSupport"], data["feature"]["sdSupport"])
		if "keyboardControl" in data["feature"]: s.setBoolean(["feature", "keyboardControl"], data["feature"]["keyboardControl"])
		if "pollWatched" in data["feature"]: s.setBoolean(["feature", "pollWatched"], data["feature"]["pollWatched"])
		if "modelSizeDetection" in data["feature"]: s.setBoolean(["feature", "modelSizeDetection"], data["feature"]["modelSizeDetection"])
		if "printStartConfirmation" in data["feature"]: s.setBoolean(["feature", "printStartConfirmation"], data["feature"]["printStartConfirmation"])
		if "printCancelConfirmation" in data["feature"]: s.setBoolean(["feature", "printCancelConfirmation"], data["feature"]["printCancelConfirmation"])
		if "g90InfluencesExtruder" in data["feature"]: s.setBoolean(["feature", "g90InfluencesExtruder"], data["feature"]["g90InfluencesExtruder"])
		if "autoUppercaseBlacklist" in data["feature"] and isinstance(data["feature"]["autoUppercaseBlacklist"], (list, tuple)): s.set(["feature", "autoUppercaseBlacklist"], data["feature"]["autoUppercaseBlacklist"])

	if "gcodeAnalysis" in data:
		if "runAt" in data["gcodeAnalysis"]: s.set(["gcodeAnalysis", "runAt"], data["gcodeAnalysis"]["runAt"])

	if "serial" in data:
		if "autoconnect" in data["serial"]: s.setBoolean(["serial", "autoconnect"], data["serial"]["autoconnect"])
		if "port" in data["serial"]: s.set(["serial", "port"], data["serial"]["port"])
		if "baudrate" in data["serial"]: s.setInt(["serial", "baudrate"], data["serial"]["baudrate"])
		if "exclusive" in data["serial"]: s.setBoolean(["serial", "exclusive"], data["serial"]["exclusive"])
		if "timeoutConnection" in data["serial"]: s.setFloat(["serial", "timeout", "connection"], data["serial"]["timeoutConnection"], min=1.0)
		if "timeoutDetection" in data["serial"]: s.setFloat(["serial", "timeout", "detection"], data["serial"]["timeoutDetection"], min=1.0)
		if "timeoutCommunication" in data["serial"]: s.setFloat(["serial", "timeout", "communication"], data["serial"]["timeoutCommunication"], min=1.0)
		if "timeoutCommunicationBusy" in data["serial"]: s.setFloat(["serial", "timeout", "communicationBusy"], data["serial"]["timeoutCommunicationBusy"], min=1.0)
		if "timeoutTemperature" in data["serial"]: s.setFloat(["serial", "timeout", "temperature"], data["serial"]["timeoutTemperature"], min=1.0)
		if "timeoutTemperatureTargetSet" in data["serial"]: s.setFloat(["serial", "timeout", "temperatureTargetSet"], data["serial"]["timeoutTemperatureTargetSet"], min=1.0)
		if "timeoutTemperatureAutoreport" in data["serial"]: s.setFloat(["serial", "timeout", "temperatureAutoreport"], data["serial"]["timeoutTemperatureAutoreport"], min=0.0)
		if "timeoutSdStatus" in data["serial"]: s.setFloat(["serial", "timeout", "sdStatus"], data["serial"]["timeoutSdStatus"], min=1.0)
		if "timeoutSdStatusAutoreport" in data["serial"]: s.setFloat(["serial", "timeout", "sdStatusAutoreport"], data["serial"]["timeoutSdStatusAutoreport"], min=0.0)
		if "timeoutBaudrateDetectionPause" in data["serial"]: s.setFloat(["serial", "timeout", "baudrateDetectionPause"], data["serial"]["timeoutBaudrateDetectionPause"], min=0.0)
		if "timeoutPositionLogWait" in data["serial"]: s.setFloat(["serial", "timeout", "positionLogWait"], data["serial"]["timeoutPositionLogWait"], min=1.0)
		if "additionalPorts" in data["serial"] and isinstance(data["serial"]["additionalPorts"], (list, tuple)): s.set(["serial", "additionalPorts"], data["serial"]["additionalPorts"])
		if "additionalBaudrates" in data["serial"] and isinstance(data["serial"]["additionalBaudrates"], (list, tuple)): s.set(["serial", "additionalBaudrates"], data["serial"]["additionalBaudrates"])
		if "longRunningCommands" in data["serial"] and isinstance(data["serial"]["longRunningCommands"], (list, tuple)): s.set(["serial", "longRunningCommands"], data["serial"]["longRunningCommands"])
		if "checksumRequiringCommands" in data["serial"] and isinstance(data["serial"]["checksumRequiringCommands"], (list, tuple)): s.set(["serial", "checksumRequiringCommands"], data["serial"]["checksumRequiringCommands"])
		if "blockedCommands" in data["serial"] and isinstance(data["serial"]["blockedCommands"], (list, tuple)): s.set(["serial", "blockedCommands"], data["serial"]["blockedCommands"])
		if "pausingCommands" in data["serial"] and isinstance(data["serial"]["pausingCommands"], (list, tuple)): s.set(["serial", "pausingCommands"], data["serial"]["pausingCommands"])
		if "emergencyCommands" in data["serial"] and isinstance(data["serial"]["emergencyCommands"], (list, tuple)): s.set(["serial", "emergencyCommands"], data["serial"]["emergencyCommands"])
		if "helloCommand" in data["serial"]: s.set(["serial", "helloCommand"], data["serial"]["helloCommand"])
		if "ignoreErrorsFromFirmware" in data["serial"]: s.setBoolean(["serial", "ignoreErrorsFromFirmware"], data["serial"]["ignoreErrorsFromFirmware"])
		if "disconnectOnErrors" in data["serial"]: s.setBoolean(["serial", "disconnectOnErrors"], data["serial"]["disconnectOnErrors"])
		if "triggerOkForM29" in data["serial"]: s.setBoolean(["serial", "triggerOkForM29"], data["serial"]["triggerOkForM29"])
		if "supportResendsWithoutOk" in data["serial"]:
			value = data["serial"]["supportResendsWithoutOk"]
			if value in ("always", "detect", "never"):
				s.set(["serial", "supportResendsWithoutOk"], value)
		if "waitForStart" in data["serial"]: s.setBoolean(["serial", "waitForStartOnConnect"], data["serial"]["waitForStart"])
		if "alwaysSendChecksum" in data["serial"]: s.setBoolean(["serial", "alwaysSendChecksum"], data["serial"]["alwaysSendChecksum"])
		if "neverSendChecksum" in data["serial"]: s.setBoolean(["serial", "neverSendChecksum"], data["serial"]["neverSendChecksum"])
		if "sdRelativePath" in data["serial"]: s.setBoolean(["serial", "sdRelativePath"], data["serial"]["sdRelativePath"])
		if "sdAlwaysAvailable" in data["serial"]: s.setBoolean(["serial", "sdAlwaysAvailable"], data["serial"]["sdAlwaysAvailable"])
		if "swallowOkAfterResend" in data["serial"]: s.setBoolean(["serial", "swallowOkAfterResend"], data["serial"]["swallowOkAfterResend"])
		if "repetierTargetTemp" in data["serial"]: s.setBoolean(["serial", "repetierTargetTemp"], data["serial"]["repetierTargetTemp"])
		if "externalHeatupDetection" in data["serial"]: s.setBoolean(["serial", "externalHeatupDetection"], data["serial"]["externalHeatupDetection"])
		if "ignoreIdenticalResends" in data["serial"]: s.setBoolean(["serial", "ignoreIdenticalResends"], data["serial"]["ignoreIdenticalResends"])
		if "firmwareDetection" in data["serial"]: s.setBoolean(["serial", "firmwareDetection"], data["serial"]["firmwareDetection"])
		if "blockWhileDwelling" in data["serial"]: s.setBoolean(["serial", "blockWhileDwelling"], data["serial"]["blockWhileDwelling"])
		if "useParityWorkaround" in data["serial"]:
			value = data["serial"]["useParityWorkaround"]
			if value in ("always", "detect", "never"):
				s.set(["serial", "useParityWorkaround"], value)
		if "sendM112OnError" in data["serial"]: s.setBoolean(["serial", "sendM112OnError"], data["serial"]["sendM112OnError"])
		if "ackMax" in data["serial"]: s.setInt(["serial", "ackMax"], data["serial"]["ackMax"])
		if "logPositionOnPause" in data["serial"]: s.setBoolean(["serial", "logPositionOnPause"], data["serial"]["logPositionOnPause"])
		if "logPositionOnCancel" in data["serial"]: s.setBoolean(["serial", "logPositionOnCancel"], data["serial"]["logPositionOnCancel"])
		if "abortHeatupOnCancel" in data["serial"]: s.setBoolean(["serial", "abortHeatupOnCancel"], data["serial"]["abortHeatupOnCancel"])
		if "maxTimeoutsIdle" in data["serial"]: s.setInt(["serial", "maxCommunicationTimeouts", "idle"], data["serial"]["maxTimeoutsIdle"])
		if "maxTimeoutsPrinting" in data["serial"]: s.setInt(["serial", "maxCommunicationTimeouts", "printing"], data["serial"]["maxTimeoutsPrinting"])
		if "maxTimeoutsLong" in data["serial"]: s.setInt(["serial", "maxCommunicationTimeouts", "long"], data["serial"]["maxTimeoutsLong"])
		if "capAutoreportTemp" in data["serial"]: s.setBoolean(["serial", "capabilities", "autoreport_temp"], data["serial"]["capAutoreportTemp"])
		if "capAutoreportSdStatus" in data["serial"]: s.setBoolean(["serial", "capabilities", "autoreport_sdstatus"], data["serial"]["capAutoreportSdStatus"])
		if "capBusyProtocol" in data["serial"]: s.setBoolean(["serial", "capabilities", "busy_protocol"], data["serial"]["capBusyProtocol"])
		if "capEmergencyParser" in data["serial"]: s.setBoolean(["serial", "capabilities", "emergency_parser"], data["serial"]["capEmergencyParser"])

		oldLog = s.getBoolean(["serial", "log"])
		if "log" in data["serial"]: s.setBoolean(["serial", "log"], data["serial"]["log"])
		if oldLog and not s.getBoolean(["serial", "log"]):
			# disable debug logging to serial.log
			logging.getLogger("SERIAL").debug("Disabling serial logging")
			logging.getLogger("SERIAL").setLevel(logging.CRITICAL)
		elif not oldLog and s.getBoolean(["serial", "log"]):
			# enable debug logging to serial.log
			logging.getLogger("SERIAL").setLevel(logging.DEBUG)
			logging.getLogger("SERIAL").debug("Enabling serial logging")

	if "temperature" in data:
		if "profiles" in data["temperature"]:
			result = []
			for profile in data["temperature"]["profiles"]:
				try:
					profile["bed"] = int(profile["bed"])
					profile["extruder"] = int(profile["extruder"])
				except ValueError:
					pass
				result.append(profile)
			s.set(["temperature", "profiles"], result)
		if "cutoff" in data["temperature"]:
			try:
				cutoff = int(data["temperature"]["cutoff"])
				if cutoff > 1:
					s.setInt(["temperature", "cutoff"], cutoff)
			except ValueError:
				pass
		if "sendAutomatically" in data["temperature"]: s.setBoolean(["temperature", "sendAutomatically"], data["temperature"]["sendAutomatically"])
		if "sendAutomaticallyAfter" in data["temperature"]: s.setInt(["temperature", "sendAutomaticallyAfter"], data["temperature"]["sendAutomaticallyAfter"], min=0, max=30)

	if "terminalFilters" in data:
		s.set(["terminalFilters"], data["terminalFilters"])

	if "system" in data:
		if "actions" in data["system"]: s.set(["system", "actions"], data["system"]["actions"])
		if "events" in data["system"]: s.set(["system", "events"], data["system"]["events"])

	if "scripts" in data:
		if "gcode" in data["scripts"] and isinstance(data["scripts"]["gcode"], dict):
			for name, script in data["scripts"]["gcode"].items():
				if name == "snippets":
					continue
				s.saveScript("gcode", name, script.replace("\r\n", "\n").replace("\r", "\n"))

	if "server" in data:
		if "commands" in data["server"]:
			if "systemShutdownCommand" in data["server"]["commands"]: s.set(["server", "commands", "systemShutdownCommand"], data["server"]["commands"]["systemShutdownCommand"])
			if "systemRestartCommand" in data["server"]["commands"]: s.set(["server", "commands", "systemRestartCommand"], data["server"]["commands"]["systemRestartCommand"])
			if "serverRestartCommand" in data["server"]["commands"]: s.set(["server", "commands", "serverRestartCommand"], data["server"]["commands"]["serverRestartCommand"])
		if "diskspace" in data["server"]:
			if "warning" in data["server"]["diskspace"]: s.setInt(["server", "diskspace", "warning"], data["server"]["diskspace"]["warning"])
			if "critical" in data["server"]["diskspace"]: s.setInt(["server", "diskspace", "critical"], data["server"]["diskspace"]["critical"])
		if "onlineCheck" in data["server"]:
			if "enabled" in data["server"]["onlineCheck"]: s.setBoolean(["server", "onlineCheck", "enabled"], data["server"]["onlineCheck"]["enabled"])
			if "interval" in data["server"]["onlineCheck"]:
				try:
					interval = int(data["server"]["onlineCheck"]["interval"])
					s.setInt(["server", "onlineCheck", "interval"], interval*60)
				except ValueError:
					pass
			if "host" in data["server"]["onlineCheck"]: s.set(["server", "onlineCheck", "host"], data["server"]["onlineCheck"]["host"])
			if "port" in data["server"]["onlineCheck"]: s.setInt(["server", "onlineCheck", "port"], data["server"]["onlineCheck"]["port"])
		if "pluginBlacklist" in data["server"]:
			if "enabled" in data["server"]["pluginBlacklist"]: s.setBoolean(["server", "pluginBlacklist", "enabled"], data["server"]["pluginBlacklist"]["enabled"])
			if "url" in data["server"]["pluginBlacklist"]: s.set(["server", "pluginBlacklist", "url"], data["server"]["pluginBlacklist"]["url"])
			if "ttl" in data["server"]["pluginBlacklist"]:
				try:
					ttl = int(data["server"]["pluginBlacklist"]["ttl"])
					s.setInt(["server", "pluginBlacklist", "ttl"], ttl * 60)
				except ValueError:
					pass
		if "allowFraming" in data["server"]:
			s.setBoolean(["server", "allowFraming"], data["server"]["allowFraming"])

	if "plugins" in data:
		for plugin in octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.SettingsPlugin):
			plugin_id = plugin._identifier
			if plugin_id in data["plugins"]:
				try:
					plugin.on_settings_save(data["plugins"][plugin_id])
				except TypeError:
					logger.warning("Could not save settings for plugin {name} ({version}) since it called super(...)".format(name=plugin._plugin_name, version=plugin._plugin_version))
					logger.warning("in a way which has issues due to OctoPrint's dynamic reloading after plugin operations.")
					logger.warning("Please contact the plugin's author and ask to update the plugin to use a direct call like")
					logger.warning("octoprint.plugin.SettingsPlugin.on_settings_save(self, data) instead.")
				except Exception:
					logger.exception("Could not save settings for plugin {name} ({version})".format(version=plugin._plugin_version,
					                                                                                name=plugin._plugin_name),
					                 extra=dict(plugin=plugin._identifier))

	s.save(trigger_event=True)



#!/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Haeussge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import errno
import sys
import traceback
import time

from past.builtins import unicode

def _log_call(*lines):
	_log(lines, prefix=">", stream="call")


def _log_stdout(*lines):
	_log(lines, prefix=" ", stream="stdout")


def _log_stderr(*lines):
	_log(lines, prefix=" ", stream="stderr")


def _log(lines, prefix=None, stream=None):
	output_stream = sys.stdout
	if stream == "stderr":
		output_stream = sys.stderr

	for line in lines:
		to_print = _to_bytes("{} {}".format(prefix, _to_unicode(line.rstrip(), errors="replace")),
		                     errors="replace")
		print(to_print, file=output_stream)


def _to_unicode(s_or_u, encoding="utf-8", errors="strict"):
	"""Make sure ``s_or_u`` is a unicode string."""
	if isinstance(s_or_u, bytes):
		return s_or_u.decode(encoding, errors=errors)
	else:
		return s_or_u


def _to_bytes(s_or_u, encoding="utf-8", errors="strict"):
	"""Make sure ``s_or_u`` is a str."""
	if isinstance(s_or_u, unicode):
		return s_or_u.encode(encoding, errors=errors)
	else:
		return s_or_u


def _execute(command, **kwargs):
	import sarge

	if isinstance(command, (list, tuple)):
		joined_command = " ".join(command)
	else:
		joined_command = command
	_log_call(joined_command)

	kwargs.update(dict(async_=True,
	                   stdout=sarge.Capture(),
	                   stderr=sarge.Capture()))

	try:
		p = sarge.run(command, **kwargs)
		while len(p.commands) == 0:
			# somewhat ugly... we can't use wait_events because
			# the events might not be all set if an exception
			# by sarge is triggered within the async process
			# thread
			time.sleep(0.01)

		# by now we should have a command, let's wait for its
		# process to have been prepared
		p.commands[0].process_ready.wait()

		if not p.commands[0].process:
			# the process might have been set to None in case of any exception
			print("Error while trying to run command {}".format(joined_command), file=sys.stderr)
			return None, [], []
	except Exception:
		print("Error while trying to run command {}".format(joined_command), file=sys.stderr)
		traceback.print_exc(file=sys.stderr)
		return None, [], []

	all_stdout = []
	all_stderr = []
	try:
		while p.commands[0].poll() is None:
			lines = p.stderr.readlines(timeout=0.5)
			if lines:
				lines = list(map(lambda x: _to_unicode(x, errors="replace"), lines))
				_log_stderr(*lines)
				all_stderr += lines

			lines = p.stdout.readlines(timeout=0.5)
			if lines:
				lines = list(map(lambda x: _to_unicode(x, errors="replace"), lines))
				_log_stdout(*lines)
				all_stdout += lines

	finally:
		p.close()

	lines = p.stderr.readlines()
	if lines:
		lines = list(map(lambda x: _to_unicode(x, errors="replace"), lines))
		_log_stderr(*lines)
		all_stderr += lines

	lines = p.stdout.readlines()
	if lines:
		lines = list(map(lambda x: _to_unicode(x, errors="replace"), lines))
		_log_stdout(*lines)
		all_stdout += lines

	return p.returncode, all_stdout, all_stderr


def _get_git_executables():
	GITS = ["git"]
	if sys.platform == "win32":
		GITS = ["git.cmd", "git.exe"]
	return GITS


def _git(args, cwd, git_executable=None):
	if git_executable is not None:
		commands = [git_executable]
	else:
		commands = _get_git_executables()

	for c in commands:
		command = [c] + args
		try:
			return _execute(command, cwd=cwd)
		except EnvironmentError:
			e = sys.exc_info()[1]
			if e.errno == errno.ENOENT:
				continue

			print("Error while trying to run command {}".format(" ".join(command)), file=sys.stderr)
			traceback.print_exc(file=sys.stderr)
			return None, [], []
		except Exception:
			print("Error while trying to run command {}".format(" ".join(command)), file=sys.stderr)
			traceback.print_exc(file=sys.stderr)
			return None, [], []
	else:
		print("Unable to find git command, tried {}".format(", ".join(commands)), file=sys.stderr)
		return None, [], []


def _python(args, cwd, python_executable, sudo=False):
	command = [python_executable] + args
	if sudo:
		command = ["sudo"] + command
	try:
		return _execute(command, cwd=cwd)
	except Exception:
		import traceback
		print("Error while trying to run command {}".format(" ".join(command)), file=sys.stderr)
		traceback.print_exc(file=sys.stderr)
		return None, [], []


def _to_error(*lines):
	if len(lines) == 1:
		if isinstance(lines[0], (list, tuple)):
			lines = lines[0]
		elif not isinstance(lines[0], (str, unicode)):
			lines = [repr(lines[0]),]
	return "\n".join(map(lambda x: _to_unicode(x, errors="replace"), lines))


def _rescue_changes(git_executable, folder):
	print(">>> Running: git diff --shortstat")
	returncode, stdout, stderr = _git(["diff", "--shortstat"], folder, git_executable=git_executable)
	if returncode is None or returncode != 0:
		raise RuntimeError("Could not update, \"git diff\" failed with returncode {}".format(returncode))
	if stdout and "".join(stdout).strip():
		# we got changes in the working tree, maybe from the user, so we'll now rescue those into a patch
		import time
		import os
		timestamp = time.strftime("%Y%m%d%H%M")
		patch = os.path.join(folder, "{}-preupdate.patch".format(timestamp))

		print(">>> Running: git diff and saving output to {}".format(patch))
		returncode, stdout, stderr = _git(["diff"], folder, git_executable=git_executable)
		if returncode is None or returncode != 0:
			raise RuntimeError("Could not update, installation directory was dirty and state could not be persisted as a patch to {}".format(patch))

		import io
		with io.open(patch, "wt", encoding="utf-8", errors="replace") as f:
			for line in stdout:
				f.write(line)

		return True

	return False


def update_source(git_executable, folder, target, force=False, branch=None):
	if _rescue_changes(git_executable, folder):
		print(">>> Running: git reset --hard")
		returncode, stdout, stderr = _git(["reset", "--hard"], folder, git_executable=git_executable)
		if returncode is None or returncode != 0:
			raise RuntimeError("Could not update, \"git reset --hard\" failed with returncode {}".format(returncode))

		print(">>> Running: git clean -f -d -e *-preupdate.patch")
		returncode, stdout, stderr = _git(["clean", "-f", "-d", "-e", "*-preupdate.patch"], folder, git_executable=git_executable)
		if returncode is None or returncode != 0:
			raise RuntimeError("Could not update, \"git clean -f\" failed with returncode {}".format(returncode))

	print(">>> Running: git fetch")
	returncode, stdout, stderr = _git(["fetch"], folder, git_executable=git_executable)
	if returncode is None or returncode != 0:
		raise RuntimeError("Could not update, \"git fetch\" failed with returncode {}".format(returncode))

	if branch is not None and branch.strip() != "":
		print(">>> Running: git checkout {}".format(branch))
		returncode, stdout, stderr = _git(["checkout", branch], folder, git_executable=git_executable)
		if returncode is None or returncode != 0:
			raise RuntimeError("Could not update, \"git checkout\" failed with returncode {}".format(returncode))

	print(">>> Running: git pull")
	returncode, stdout, stderr = _git(["pull"], folder, git_executable=git_executable)
	if returncode is None or returncode != 0:
		raise RuntimeError("Could not update, \"git pull\" failed with returncode {}".format(returncode))

	if force:
		reset_command = ["reset", "--hard"]
		reset_command += [target]

		print(">>> Running: git {}".format(" ".join(reset_command)))
		returncode, stdout, stderr = _git(reset_command, folder, git_executable=git_executable)
		if returncode is None or returncode != 0:
			raise RuntimeError("Error while updating, \"git {}\" failed with returncode {}".format(" ".join(reset_command), returncode))


def install_source(python_executable, folder, user=False, sudo=False):
	print(">>> Running: python setup.py clean")
	returncode, stdout, stderr = _python(["setup.py", "clean"], folder, python_executable)
	if returncode is None or returncode != 0:
		print("\"python setup.py clean\" failed with returncode {}".format(returncode))
		print("Continuing anyways")

	print(">>> Running: python setup.py install")
	args = ["setup.py", "install"]
	if user:
		args.append("--user")
	returncode, stdout, stderr = _python(args, folder, python_executable, sudo=sudo)
	if returncode is None or returncode != 0:
		raise RuntimeError("Could not update, \"python setup.py install\" failed with returncode {}".format(returncode))


def parse_arguments():
	import argparse

	boolean_trues = ["true", "yes", "1"]

	parser = argparse.ArgumentParser(prog="update-octoprint.py")

	parser.add_argument("--git", action="store", type=str, dest="git_executable",
	                    help="Specify git executable to use")
	parser.add_argument("--python", action="store", type=str, dest="python_executable",
	                    help="Specify python executable to use")
	parser.add_argument("--force", action="store", type=lambda x: x in boolean_trues,
	                    dest="force", default=False,
	                    help="Set this to true to force the update to only the specified version (nothing newer, nothing older)")
	parser.add_argument("--sudo", action="store_true", dest="sudo",
	                    help="Install with sudo")
	parser.add_argument("--user", action="store_true", dest="user",
	                    help="Install to the user site directory instead of the general site directory")
	parser.add_argument("--branch", action="store", type=str, dest="branch", default=None,
	                    help="Specify the branch to make sure is checked out")
	parser.add_argument("folder", type=str,
	                    help="Specify the base folder of the OctoPrint installation to update")
	parser.add_argument("target", type=str,
	                    help="Specify the commit or tag to which to update")

	args = parser.parse_args()

	return args

def main():
	args = parse_arguments()

	git_executable = None
	if args.git_executable:
		git_executable = args.git_executable

	python_executable = sys.executable
	if args.python_executable:
		python_executable = args.python_executable
		if python_executable.startswith('"'):
			python_executable = python_executable[1:]
		if python_executable.endswith('"'):
			python_executable = python_executable[:-1]

	print("Python executable: {!r}".format(python_executable))

	folder = args.folder

	import os
	if not os.access(folder, os.W_OK):
		raise RuntimeError("Could not update, base folder is not writable")

	update_source(git_executable, folder, args.target, force=args.force, branch=args.branch)
	install_source(python_executable, folder, user=args.user, sudo=args.sudo)

if __name__ == "__main__":
	main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import click
click.disable_unicode_literals_warning = True
import octoprint
import sys

#~~ click context

class OctoPrintContext(object):
	"""Custom context wrapping the standard options."""

	def __init__(self, configfile=None, basedir=None, verbosity=0, safe_mode=False):
		self.configfile = configfile
		self.basedir = basedir
		self.verbosity = verbosity
		self.safe_mode = safe_mode

pass_octoprint_ctx = click.make_pass_decorator(OctoPrintContext, ensure=True)
"""Decorator to pass in the :class:`OctoPrintContext` instance."""

#~~ Custom click option to hide from help

class HiddenOption(click.Option):
	"""Custom option sub class with empty help."""
	def get_help_record(self, ctx):
		pass

def hidden_option(*param_decls, **attrs):
	"""Attaches a hidden option to the command.  All positional arguments are
	passed as parameter declarations to :class:`Option`; all keyword
	arguments are forwarded unchanged.  This is equivalent to creating an
	:class:`Option` instance manually and attaching it to the
	:attr:`Command.params` list.
	"""

	import inspect
	from click.decorators import _param_memo

	def decorator(f):
		if 'help' in attrs:
			attrs['help'] = inspect.cleandoc(attrs['help'])
		_param_memo(f, HiddenOption(param_decls, **attrs))
		return f
	return decorator

#~~ helper for setting context options

def set_ctx_obj_option(ctx, param, value):
	"""Helper for setting eager options on the context."""
	if ctx.obj is None:
		ctx.obj = OctoPrintContext()
	if value != param.default:
		setattr(ctx.obj, param.name, value)
	elif param.default is not None:
		setattr(ctx.obj, param.name, param.default)

#~~ helper for retrieving context options

def get_ctx_obj_option(ctx, key, default, include_parents=True):
	if include_parents and hasattr(ctx, "parent") and ctx.parent:
		fallback = get_ctx_obj_option(ctx.parent, key, default)
	else:
		fallback = default
	return getattr(ctx.obj, key, fallback)

#~~ helper for setting a lot of bulk options

def bulk_options(options):
	"""
	Utility decorator to decorate a function with a list of click decorators.

	The provided list of ``options`` will be reversed to ensure correct
	processing order (inverse from what would be intuitive).
	"""

	def decorator(f):
		options.reverse()
		for option in options:
			option(f)
		return f
	return decorator

#~~ helper for setting --basedir, --config and --verbose options

def standard_options(hidden=False):
	"""
	Decorator to add the standard options shared among all "octoprint" commands.

	Adds the options ``--basedir``, ``--config`` and ``--verbose``. If ``hidden``
	is set to ``True``, the options will be available on the command but not
	listed in its help page.
	"""

	factory = click.option
	if hidden:
		factory = hidden_option

	options = [
		factory("--basedir", "-b", type=click.Path(), callback=set_ctx_obj_option, is_eager=True, expose_value=False,
		        help="Specify the basedir to use for configs, uploads, timelapses etc."),
		factory("--config", "-c", "configfile", type=click.Path(), callback=set_ctx_obj_option, is_eager=True, expose_value=False,
		        help="Specify the config file to use."),
		factory("--verbose", "-v", "verbosity", count=True, callback=set_ctx_obj_option, is_eager=True, expose_value=False,
		        help="Increase logging verbosity."),
		factory("--safe", "safe_mode", is_flag=True, callback=set_ctx_obj_option, is_eager=True, expose_value=False,
		        help="Enable safe mode; disables all third party plugins.")
	]

	return bulk_options(options)

#~~ helper for settings legacy options we still have to support on "octoprint"

legacy_options = bulk_options([
	hidden_option("--host", type=click.STRING, callback=set_ctx_obj_option),
	hidden_option("--port", type=click.INT, callback=set_ctx_obj_option),
	hidden_option("--logging", type=click.Path(), callback=set_ctx_obj_option),
	hidden_option("--debug", "-d", is_flag=True, callback=set_ctx_obj_option),
	hidden_option("--daemon", type=click.Choice(["start", "stop", "restart"]), callback=set_ctx_obj_option),
	hidden_option("--pid", type=click.Path(), default="/tmp/octoprint.pid", callback=set_ctx_obj_option),
	hidden_option("--iknowwhatimdoing", "allow_root", is_flag=True, callback=set_ctx_obj_option),
	hidden_option("--ignore-blacklist", "ignore_blacklist", is_flag=True, callback=set_ctx_obj_option)
])
"""Legacy options available directly on the "octoprint" command in earlier versions.
   Kept available for reasons of backwards compatibility, but hidden from the
   generated help pages."""

#~~ "octoprint" command, merges server_commands and plugin_commands groups

from .server import server_commands
from .plugins import plugin_commands
from .dev import dev_commands
from .client import client_commands
from .config import config_commands
from .analysis import analysis_commands
from .user import user_commands

@click.group(name="octoprint", invoke_without_command=True, cls=click.CommandCollection,
             sources=[server_commands, plugin_commands, dev_commands, client_commands, config_commands, analysis_commands, user_commands])
@standard_options()
@legacy_options
@click.version_option(version=octoprint.__version__, allow_from_autoenv=False)
@click.pass_context
def octo(ctx, **kwargs):

	if ctx.invoked_subcommand is None:
		# We have to support calling the octoprint command without any
		# sub commands to remain backwards compatible.
		#
		# But better print a message to inform people that they should
		# use the sub commands instead.

		def get_value(key):
			return get_ctx_obj_option(ctx, key, kwargs.get(key))
		daemon = get_value("daemon")

		if daemon:
			click.echo("Daemon operation via \"octoprint --daemon "
			           "start|stop|restart\" is deprecated, please use "
			           "\"octoprint daemon start|stop|restart\" from now on")

			if sys.platform == "win32" or sys.platform == "darwin":
				click.echo("Sorry, daemon mode is not supported under your operating system right now")
			else:
				from octoprint.cli.server import daemon_command
				ctx.invoke(daemon_command, command=daemon, **kwargs)
		else:
			click.echo("Starting the server via \"octoprint\" is deprecated, "
			           "please use \"octoprint serve\" from now on.")

			from octoprint.cli.server import serve_command
			ctx.invoke(serve_command, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os, struct, sys, time

from serial import Serial
from serial import SerialException, SerialTimeoutException
from builtins import range

from . import ispBase, intelHex

class Stk500v2(ispBase.IspBase):
	def __init__(self):
		self.serial = None
		self.seq = 1
		self.lastAddr = -1
		self.progressCallback = None
	
	def connect(self, port = 'COM22', speed = 115200):
		if self.serial is not None:
			self.close()
		try:
			self.serial = Serial(str(port), speed, timeout=1, writeTimeout=10000)
		except SerialException:
			raise ispBase.IspError("Failed to open serial port")
		except Exception:
			raise ispBase.IspError("Unexpected error while connecting to serial port:" + port + ":" + str(sys.exc_info()[0]))
		self.seq = 1
		
		#Reset the controller
		self.serial.setDTR(1)
		time.sleep(0.1)
		self.serial.setDTR(0)
		time.sleep(0.2)
		
		self.sendMessage([1])
		if self.sendMessage([0x10, 0xc8, 0x64, 0x19, 0x20, 0x00, 0x53, 0x03, 0xac, 0x53, 0x00, 0x00]) != [0x10, 0x00]:
			self.close()
			raise ispBase.IspError("Failed to enter programming mode")

	def close(self):
		if self.serial is not None:
			self.serial.close()
			self.serial = None

	#Leave ISP does not reset the serial port, only resets the device, and returns the serial port after disconnecting it from the programming interface.
	#	This allows you to use the serial port without opening it again.
	def leaveISP(self):
		if self.serial is not None:
			if self.sendMessage([0x11]) != [0x11, 0x00]:
				raise ispBase.IspError("Failed to leave programming mode")
			ret = self.serial
			self.serial = None
			return ret
		return None
	
	def isConnected(self):
		return self.serial is not None
	
	def sendISP(self, data):
		recv = self.sendMessage([0x1D, 4, 4, 0, data[0], data[1], data[2], data[3]])
		return recv[2:6]
	
	def writeFlash(self, flashData):
		#Set load addr to 0, in case we have more then 64k flash we need to enable the address extension
		pageSize = self.chip['pageSize'] * 2
		flashSize = pageSize * self.chip['pageCount']
		if flashSize > 0xFFFF:
			self.sendMessage([0x06, 0x80, 0x00, 0x00, 0x00])
		else:
			self.sendMessage([0x06, 0x00, 0x00, 0x00, 0x00])
		
		loadCount = (len(flashData) + pageSize - 1) // pageSize
		for i in range(0, loadCount):
			self.sendMessage([0x13, pageSize >> 8, pageSize & 0xFF, 0xc1, 0x0a, 0x40, 0x4c, 0x20, 0x00, 0x00] + flashData[(i * pageSize):(i * pageSize + pageSize)])
			if self.progressCallback is not None:
				self.progressCallback(i + 1, loadCount*2)
	
	def verifyFlash(self, flashData):
		#Set load addr to 0, in case we have more then 64k flash we need to enable the address extension
		flashSize = self.chip['pageSize'] * 2 * self.chip['pageCount']
		if flashSize > 0xFFFF:
			self.sendMessage([0x06, 0x80, 0x00, 0x00, 0x00])
		else:
			self.sendMessage([0x06, 0x00, 0x00, 0x00, 0x00])
		
		loadCount = (len(flashData) + 0xFF) // 0x100
		for i in range(0, loadCount):
			recv = self.sendMessage([0x14, 0x01, 0x00, 0x20])[2:0x102]
			if self.progressCallback is not None:
				self.progressCallback(loadCount + i + 1, loadCount*2)
			for j in range(0, 0x100):
				if i * 0x100 + j < len(flashData) and flashData[i * 0x100 + j] != recv[j]:
					raise ispBase.IspError('Verify error at: 0x%x' % (i * 0x100 + j))

	def sendMessage(self, data):
		message = struct.pack(">BBHB", 0x1B, self.seq, len(data), 0x0E)
		for c in data:
			message += struct.pack(">B", c)
		checksum = 0
		for c in bytearray(message):
			checksum ^= c
		message += struct.pack(">B", checksum)
		try:
			self.serial.write(message)
			self.serial.flush()
		except SerialTimeoutException:
			raise ispBase.IspError('Serial send timeout')
		self.seq = (self.seq + 1) & 0xFF
		return self.recvMessage()
	
	def recvMessage(self):
		state = 'Start'
		checksum = 0
		while True:
			for _ in range(0, 5):
				s = self.serial.read()
				if len(s) > 0:
					break
			else:
				raise ispBase.IspError("Timeout")
			b = struct.unpack(">B", s)[0]
			checksum ^= b
			#print(hex(b))
			if state == 'Start':
				if b == 0x1B:
					state = 'GetSeq'
					checksum = 0x1B
			elif state == 'GetSeq':
				state = 'MsgSize1'
			elif state == 'MsgSize1':
				msgSize = b << 8
				state = 'MsgSize2'
			elif state == 'MsgSize2':
				msgSize |= b
				state = 'Token'
			elif state == 'Token':
				if b != 0x0E:
					state = 'Start'
				else:
					state = 'Data'
					data = []
			elif state == 'Data':
				data.append(b)
				if len(data) == msgSize:
					state = 'Checksum'
			elif state == 'Checksum':
				if checksum != 0:
					state = 'Start'
				else:
					return data


def main():
	programmer = Stk500v2()
	programmer.connect(port = sys.argv[1])
	programmer.programChip(intelHex.readHex(sys.argv[2]))
	sys.exit(1)

if __name__ == '__main__':
	main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

ADMIN_GROUP = "admins"
USER_GROUP = "users"
GUEST_GROUP = "guests"
READONLY_GROUP = "readonly"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
# MO unicode_literals because fixing all the doctests is too annoying :-P

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import requests
import logging

from octoprint.util import sv
from octoprint.util.version import get_comparable_version

RELEASE_URL = "https://api.github.com/repos/{user}/{repo}/releases"

logger = logging.getLogger("octoprint.plugins.softwareupdate.version_checks.github_release")

def _filter_out_latest(releases,
                       sort_key=None,
                       include_prerelease=False,
                       commitish=None):
	"""
	Filters out the newest of all matching releases.

	Tests:

	    >>> release_1_2_15 = dict(name="1.2.15", tag_name="1.2.15", html_url="some_url", published_at="2016-07-29T19:53:29Z", prerelease=False, draft=False, target_commitish="prerelease")
	    >>> release_1_2_16rc1 = dict(name="1.2.16rc1", tag_name="1.2.16rc1", html_url="some_url", published_at="2016-08-29T12:00:00Z", prerelease=True, draft=False, target_commitish="rc/maintenance")
	    >>> release_1_2_16rc2 = dict(name="1.2.16rc2", tag_name="1.2.16rc2", html_url="some_url", published_at="2016-08-30T12:00:00Z", prerelease=True, draft=False, target_commitish="rc/maintenance")
	    >>> release_1_2_17rc1 = dict(name="1.2.17rc1", tag_name="1.2.17rc1", html_url="some_url", published_at="2016-08-31T12:00:00Z", prerelease=True, draft=True, target_commitish="rc/maintenance")
	    >>> release_1_3_0rc1 = dict(name="1.3.0rc1", tag_name="1.3.0rc1", html_url="some_url", published_at="2016-12-12T12:00:00Z", prerelease=True, draft=False, target_commitish="rc/devel")
	    >>> release_1_3_5rc1 = dict(name="1.3.5rc1", tag_name="1.3.5rc1", html_url="some_url", published_at="2017-06-14T10:00:00Z", prerelease=True, draft=False, target_commitish="rc/maintenance")
	    >>> release_1_2_18 = dict(name="1.2.18", tag_name="1.2.18", html_url="some_url", published_at="2016-12-13T12:00:00Z", prerelease=False, draft=False, target_commitish="master")
	    >>> release_1_4_0rc1 = dict(name="1.4.0rc1", tag_name="1.4.0rc1", html_url="some_url", published_at="2017-12-12T12:00:00Z", prerelease=True, draft=False, target_commitish="rc/future")
	    >>> release_1_4_0rc1_devel = dict(name="1.4.0rc1", tag_name="1.4.0rc1", html_url="some_url", published_at="2017-12-12T12:00:00Z", prerelease=True, draft=False, target_commitish="rc/devel")
	    >>> releases = [release_1_2_15, release_1_2_16rc1, release_1_2_16rc2, release_1_2_17rc1, release_1_3_0rc1, release_1_4_0rc1]
	    >>> _filter_out_latest(releases, include_prerelease=False, commitish=None)
	    ('1.2.15', '1.2.15', 'some_url')
	    >>> _filter_out_latest(releases, include_prerelease=True, commitish=["rc/maintenance"])
	    ('1.2.16rc2', '1.2.16rc2', 'some_url')
	    >>> _filter_out_latest(releases, include_prerelease=True, commitish=["rc/devel"])
	    ('1.3.0rc1', '1.3.0rc1', 'some_url')
	    >>> _filter_out_latest(releases, include_prerelease=True, commitish=None)
	    ('1.4.0rc1', '1.4.0rc1', 'some_url')
	    >>> _filter_out_latest(releases, include_prerelease=True, commitish=["rc/doesntexist"])
	    ('1.2.15', '1.2.15', 'some_url')
	    >>> _filter_out_latest([release_1_2_17rc1])
	    (None, None, None)
	    >>> _filter_out_latest([release_1_2_16rc1, release_1_2_16rc2])
	    (None, None, None)

	    >>> comparable_factory = _get_comparable_factory("python", force_base=True)
	    >>> sort_key = lambda release: comparable_factory(_get_sanitized_version(release["tag_name"]))
	    >>> _filter_out_latest(releases + [release_1_2_18], include_prerelease=False, commitish=None, sort_key=sort_key)
	    ('1.2.18', '1.2.18', 'some_url')
	    >>> _filter_out_latest(releases + [release_1_2_18], include_prerelease=True, commitish=["rc/maintenance"], sort_key=sort_key)
	    ('1.2.18', '1.2.18', 'some_url')
	    >>> _filter_out_latest(releases + [release_1_2_18], include_prerelease=True, commitish=["rc/devel"], sort_key=sort_key)
	    ('1.3.0rc1', '1.3.0rc1', 'some_url')

	    >>> _filter_out_latest([release_1_2_18, release_1_3_5rc1], include_prerelease=True, commitish=["rc/maintenance"])
	    ('1.3.5rc1', '1.3.5rc1', 'some_url')
	    >>> _filter_out_latest([release_1_2_18, release_1_3_5rc1], include_prerelease=True, commitish=["rc/maintenance", "rc/devel"])
	    ('1.3.5rc1', '1.3.5rc1', 'some_url')
	    >>> _filter_out_latest([release_1_2_18, release_1_3_5rc1, release_1_4_0rc1_devel], include_prerelease=True, commitish=["rc/maintenance"])
	    ('1.3.5rc1', '1.3.5rc1', 'some_url')
	    >>> _filter_out_latest([release_1_2_18, release_1_3_5rc1, release_1_4_0rc1_devel], include_prerelease=True, commitish=["rc/maintenance", "rc/devel"])
	    ('1.4.0rc1', '1.4.0rc1', 'some_url')
	"""

	nothing = None, None, None

	if sort_key is None:
		sort_key = lambda release: sv(release.get("published_at", None))

	# filter out prereleases and drafts
	filter_function = lambda rel: not rel["prerelease"] and not rel["draft"]
	if include_prerelease:
		if commitish:
			filter_function = lambda rel: not rel["draft"] and \
			                              (not rel["prerelease"] or rel["target_commitish"] in commitish)
		else:
			filter_function = lambda rel: not rel["draft"]

	releases = list(filter(filter_function, releases))
	if not releases:
		return nothing

	# sort by sort_key
	releases = sorted(releases, key=sort_key)

	# latest release = last in list
	latest = releases[-1]

	return latest["name"], latest["tag_name"], latest.get("html_url", None)


def _get_latest_release(user, repo, compare_type,
                        include_prerelease=False,
                        commitish=None,
                        force_base=True):
	from ..exceptions import NetworkError

	nothing = None, None, None

	try:
		r = requests.get(RELEASE_URL.format(user=user, repo=repo), timeout=(3.05, 30))
	except requests.ConnectionError as exc:
		raise NetworkError(cause=exc)

	from . import log_github_ratelimit
	log_github_ratelimit(logger, r)

	if not r.status_code == requests.codes.ok:
		return nothing

	releases = r.json()

	# sanitize
	required_fields = {"name", "tag_name", "html_url", "draft", "prerelease", "published_at", "target_commitish"}
	releases = list(filter(lambda rel: set(rel.keys()) & required_fields == required_fields,
	                       releases))

	comparable_factory = _get_comparable_factory(compare_type,
	                                             force_base=force_base)
	sort_key = lambda release: comparable_factory(_get_sanitized_version(release["tag_name"]))

	return _filter_out_latest(releases,
	                          sort_key=sort_key,
	                          include_prerelease=include_prerelease,
	                          commitish=commitish)


def _get_sanitized_version(version_string):
	"""
	Removes "-..." prefix from version strings.

	Tests:
	    >>> _get_sanitized_version(None)
	    >>> _get_sanitized_version("1.2.15")
	    '1.2.15'
	    >>> _get_sanitized_version("1.2.15-dev12")
	    '1.2.15'
	"""

	if version_string is not None and "-" in version_string:
		version_string = version_string[:version_string.find("-")]
	return version_string


def _get_base_from_version_tuple(version_tuple):
	"""
	Reduces version tuple to base version.

	Tests:

	    >>> _get_base_from_version_tuple(("1", "2", "15"))
	    ('1', '2', '15')
	    >>> _get_base_from_version_tuple(("1", "2", "15", "*", "dev12"))
	    ('1', '2', '15')
	"""

	base_version = []
	for part in version_tuple:
		if part.startswith("*"):
			break
		base_version.append(part)
	return tuple(base_version)


def _get_comparable_version_semantic(version_string, force_base=True):
	import semantic_version

	version = semantic_version.Version.coerce(version_string, partial=False)

	if force_base:
		version_string = "{}.{}.{}".format(version.major, version.minor, version.patch)
		version = semantic_version.Version.coerce(version_string, partial=False)

	return version


def _get_sanitized_compare_type(compare_type, custom=None):
	if not compare_type in ("python", "python_unequal",
	                        "semantic", "semantic_unequal",
	                        "unequal", "custom") or compare_type == "custom" and custom is None:
		compare_type = "python"
	return compare_type


def _get_comparable_factory(compare_type, force_base=True):
	if compare_type in ("python", "python_unequal"):
		return lambda version: get_comparable_version(version, base=force_base)
	elif compare_type in ("semantic", "semantic_unequal"):
		return lambda version: _get_comparable_version_semantic(version, force_base=force_base)
	else:
		return lambda version: version


def _get_comparator(compare_type, custom=None):
	if compare_type in ("python", "semantic"):
		return lambda a, b: a >= b
	elif compare_type == "custom":
		return custom
	else:
		return lambda a, b: a == b


def _is_current(release_information, compare_type, custom=None, force_base=True):
	"""
	Checks if the provided release information indicates the version being the most current one.

	Tests:

	    >>> _is_current(dict(remote=dict(value=None)), "python")
	    True
	    >>> _is_current(dict(local=dict(value="1.2.15"), remote=dict(value="1.2.16")), "python")
	    False
	    >>> _is_current(dict(local=dict(value="1.2.16dev1"), remote=dict(value="1.2.16dev2")), "python")
	    True
	    >>> _is_current(dict(local=dict(value="1.2.16dev1"), remote=dict(value="1.2.16dev2")), "python", force_base=False)
	    False
	    >>> _is_current(dict(local=dict(value="1.2.16dev3"), remote=dict(value="1.2.16dev2")), "python", force_base=False)
	    True
	    >>> _is_current(dict(local=dict(value="1.2.16dev3"), remote=dict(value="1.2.16dev2")), "python_unequal", force_base=False)
	    False
	    >>> _is_current(dict(local=dict(value="1.3.0.post1+g1014712"), remote=dict(value="1.3.0")), "python")
	    True

	"""

	if release_information["remote"]["value"] is None:
		return True

	compare_type = _get_sanitized_compare_type(compare_type, custom=custom)
	comparable_factory = _get_comparable_factory(compare_type, force_base=force_base)
	comparator = _get_comparator(compare_type, custom=custom)

	sanitized_local = _get_sanitized_version(release_information["local"]["value"])
	sanitized_remote = _get_sanitized_version(release_information["remote"]["value"])

	try:
		return comparator(comparable_factory(sanitized_local),
		                  comparable_factory(sanitized_remote))
	except Exception:
		logger.exception("Could not check if version is current due to an error, assuming it is")
		return True


def get_latest(target, check, custom_compare=None, online=True):
	from ..exceptions import ConfigurationInvalid

	user = check.get("user", None)
	repo = check.get("repo", None)
	current = check.get("current", None)
	if user is None or repo is None or current is None:
		raise ConfigurationInvalid("Update configuration for {} of type github_release needs all of user, repo and current set and not None".format(target))

	information =dict(
		local=dict(name=current, value=current),
		remote=dict(name="?", value="?", release_notes=None),
		needs_online=not check.get("offline", False)
	)
	if not online and information["needs_online"]:
		return information, True

	include_prerelease = check.get("prerelease", False)
	prerelease_channel = check.get("prerelease_channel", None)

	# determine valid "commitish" values in case we track prereleases
	commitish = None
	if prerelease_channel:
		prerelease_branches = check.get("prerelease_branches", None)
		if prerelease_branches:
			# fetch valid commitish list from configured prerelease_branches for selected channel
			commitishes = dict((x["branch"], x.get("commitish", [x["branch"]])) for x in prerelease_branches)
			commitish = commitishes.get(prerelease_channel, [prerelease_channel])

	force_base = check.get("force_base", False)
	compare_type = _get_sanitized_compare_type(check.get("release_compare", "python"),
	                                           custom=custom_compare)

	remote_name, remote_tag, release_notes = _get_latest_release(check["user"],
	                                                             check["repo"],
	                                                             compare_type,
	                                                             include_prerelease=include_prerelease,
	                                                             commitish=commitish,
	                                                             force_base=force_base)

	if remote_name is None:
		if remote_tag is not None:
			remote_name = remote_tag
		else:
			remote_name = "-"

	information["remote"] = dict(name=remote_name, value=remote_tag, release_notes=release_notes)

	logger.debug("Target: %s, local: %s, remote: %s" % (target, current, remote_tag))

	return information, _is_current(information,
	                                compare_type,
	                                custom=custom_compare,
	                                force_base=force_base)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Marc Hannappel <salandora@gmail.com>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import logging
import os
from functools import partial

import yaml

from past.builtins import basestring

from octoprint.settings import settings
from octoprint.util import atomic_write
from octoprint.access import ADMIN_GROUP, USER_GROUP, GUEST_GROUP, READONLY_GROUP
from octoprint.access.permissions import Permissions, OctoPrintPermission
from octoprint.vendor.flask_principal import Need, Permission


GroupNeed = partial(Need, 'group')
GroupNeed.__doc__ = """A need with the method preset to `"group"`."""


class GroupPermission(Permission):
	def __init__(self, key):
		need = GroupNeed(key)
		super(GroupPermission, self).__init__(need)


class GroupManager(object):
	@classmethod
	def default_permissions_for_group(cls, group):
		result = []
		for permission in Permissions.all():
			if group in permission.default_groups:
				result.append(permission)
		return result

	def __init__(self):
		self._logger = logging.getLogger(__name__)
		self._group_change_listeners = []
		self._init_defaults()

	@property
	def groups(self):
		return []

	@property
	def admin_group(self):
		return self.find_group(ADMIN_GROUP)

	@property
	def user_group(self):
		return self.find_group(USER_GROUP)

	@property
	def guest_group(self):
		return self.find_group(GUEST_GROUP)

	def _init_defaults(self):
		self.add_group(ADMIN_GROUP,
		               "Admins",
		               "Administrators",
		               self.default_permissions_for_group(ADMIN_GROUP),
		               [],
		               changeable=False,
		               removable=False,
		               save=False)
		self.add_group(USER_GROUP,
		               "Users",
		               "All logged in users",
		               self.default_permissions_for_group(USER_GROUP),
		               [],
		               default=True,
		               removable=False,
		               toggleable=False,
		               save=False)
		self.add_group(GUEST_GROUP,
		               "Guests",
		               "Anyone who is not currently logged in",
		               self.default_permissions_for_group(GUEST_GROUP),
		               [],
		               removable=False,
		               toggleable=False,
		               save=False),
		self.add_group(READONLY_GROUP,
		               "Read-only Access",
		               "Group to gain read-only access",
		               self.default_permissions_for_group(READONLY_GROUP),
		               [],
		               changeable=False,
		               removable=False,
		               save=False)

	def register_listener(self, listener):
		self._group_change_listeners.append(listener)

	def unregister_listener(self, listener):
		self._group_change_listeners.remove(listener)

	def add_group(self, key, name, description, permissions, subgroups,
	              default=False, removable=True, changeable=True, toggleable=True, save=True, notify=True):
		pass

	def update_group(self, key, description=None, permissions=None, subgroups=None, default=None, save=True, notify=True):
		pass

	def remove_group(self, key, save=True, notify=True):
		pass

	def find_group(self, key):
		return None

	def _to_permissions(self, *permissions):
		return list(filter(lambda x: x is not None,
		                   [Permissions.find(permission) for permission in permissions]))

	def _from_permissions(self, *permissions):
		return [permission.key for permission in permissions]

	def _from_groups(self, *groups):
		return [group.key for group in groups]

	def _to_groups(self, *groups):
		return list(filter(lambda x: x is not None,
		                   [self._to_group(g) for g in groups]))

	def _to_group(self, group):
		# noinspection PyCompatibility
		if isinstance(group, Group):
			return group
		elif isinstance(group, basestring):
			return self.find_group(group)
		elif isinstance(group, dict):
			return self.find_group(group.get("key"))
		else:
			return None

	def _notify_listeners(self, action, group, *args, **kwargs):
		method = "on_group_{}".format(action)
		for listener in self._group_change_listeners:
			try:
				getattr(listener, method)(group, *args, **kwargs)
			except Exception:
				self._logger.exception("Error notifying listener {!r} via {}".format(listener, method))

class GroupChangeListener(object):

	def on_group_added(self, group):
		pass

	def on_group_removed(self, group):
		pass

	def on_group_permissions_changed(self, group, added=None, removed=None):
		pass

	def on_group_subgroups_changed(self, group, added=None, removed=None):
		pass


class FilebasedGroupManager(GroupManager):
	def __init__(self, path=None):
		if path is None:
			path = settings().get(["accessControl", "groupfile"])
			if path is None:
				path = os.path.join(settings().getBaseFolder("base"), "groups.yaml")

		self._groupfile = path
		self._groups = dict()
		self._dirty = False

		GroupManager.__init__(self)

		self._load()

	def _load(self):
		if os.path.exists(self._groupfile) and os.path.isfile(self._groupfile):
			try:
				with io.open(self._groupfile, 'rt', encoding='utf-8') as f:
					data = yaml.safe_load(f)

				if not "groups" in data:
					groups = data
					data = dict(groups=groups)

				groups = data.get("groups", dict())
				tracked_permissions = data.get("tracked", list())

				for key, attributes in groups.items():
					if key in self._groups:
						# group is already there (from the defaults most likely)
						if not self._groups[key].is_changeable():
							# group may not be changed -> bail
							continue

						removable = self._groups[key].is_removable()
						changeable = self._groups[key].is_changeable()
						toggleable = self._groups[key].is_toggleable()
					else:
						removable = True
						changeable = True
						toggleable = True

					permissions = self._to_permissions(*attributes.get("permissions", []))
					default_permissions = self.default_permissions_for_group(key)
					for permission in default_permissions:
						if not permission.key in tracked_permissions and not permission in permissions:
							permissions.append(permission)

					subgroups = attributes.get("subgroups", [])

					group = Group(key, attributes.get("name", ""),
					              description=attributes.get("description", ""),
					              permissions=permissions,
					              subgroups=subgroups,
					              default=attributes.get("default", False),
					              removable=removable,
					              changeable=changeable,
					              toggleable=toggleable)
					self._groups[key] = group

				for group in self._groups.values():
					group._subgroups = self._to_groups(*group._subgroups)

			except Exception:
				self._logger.exception("Error while loading groups from file {}".format(self._groupfile))

	def _save(self, force=False):
		if self._groupfile is None or not self._dirty and not force:
			return

		groups = dict()
		for key in self._groups.keys():
			group = self._groups[key]
			groups[key] = dict(
				name=group._name,
				description=group._description,
				permissions=self._from_permissions(*group._permissions),
				subgroups=self._from_groups(*group._subgroups),
				default=group._default
			)

		data = dict(groups=groups,
		            tracked=[x.key for x in Permissions.all()])

		with atomic_write(self._groupfile, mode='wt', permissions=0o600, max_permissions=0o666) as f:
			import yaml
			yaml.safe_dump(data, f, default_flow_style=False, indent=4, allow_unicode=True)
			self._dirty = False
		self._load()

	@property
	def groups(self):
		return list(self._groups.values())

	@property
	def default_groups(self):
		return [group for group in self._groups.values() if group.is_default()]

	def find_group(self, key):
		if key is None:
			return None
		return self._groups.get(key)

	def add_group(self, key, name, description, permissions, subgroups, default=False, removable=True,
	              changeable=True, toggleable=True, overwrite=False, notify=True, save=True):
		if key in self._groups and not overwrite:
			raise GroupAlreadyExists(key)

		if not permissions:
			permissions = []

		permissions = self._to_permissions(*permissions)
		assert(all(map(lambda p: isinstance(p, OctoPrintPermission), permissions)))

		subgroups = self._to_groups(*subgroups)
		assert(all(map(lambda g: isinstance(g, Group), subgroups)))

		group = Group(key, name,
		              description=description,
		              permissions=permissions,
		              subgroups=subgroups,
		              default=default,
		              changeable=changeable,
		              removable=removable,
		              toggleable=toggleable)
		self._groups[key] = group

		if save:
			self._dirty = True
			self._save()

		if notify:
			self._notify_listeners("added", group)

	def remove_group(self, key, save=True, notify=True):
		"""Removes a Group by key"""
		group = self._to_group(key)
		if group is None:
			raise UnknownGroup(key)

		if not group.is_removable():
			raise GroupUnremovable(key)

		del self._groups[key]
		self._dirty = True

		if save:
			self._save()

		if notify:
			self._notify_listeners("removed", group)

	def update_group(self, key, description=None, permissions=None, subgroups=None, default=None, save=True, notify=True):
		group = self._to_group(key)
		if group is None:
			raise UnknownGroup(key)

		if not group.is_changeable():
			raise GroupCantBeChanged(key)

		if description is not None and description != group.get_description():
			group.change_description(description)
			self._dirty = True

		notifications = []

		if permissions is not None:
			permissions = self._to_permissions(*permissions)
			assert (all(map(lambda p: isinstance(p, OctoPrintPermission), permissions)))

			removed_permissions = list(set(group._permissions) - set(permissions))
			added_permissions = list(set(permissions) - set(group._permissions))

			if removed_permissions:
				self._dirty |= group.remove_permissions_from_group(removed_permissions)
			if added_permissions:
				self._dirty |= group.add_permissions_to_group(added_permissions)

			notifications.append((("permissions_changed", group),
			                      dict(added=added_permissions, removed=removed_permissions)))

		if subgroups is not None:
			subgroups = self._to_groups(*subgroups)
			assert(all(map(lambda g: isinstance(g, Group), subgroups)))

			removed_subgroups = list(set(group._subgroups) - set(subgroups))
			added_subgroups = list(set(subgroups) - set(group._subgroups))

			if removed_subgroups:
				self._dirty = group.remove_subgroups_from_group(removed_subgroups)
			if added_subgroups:
				self._dirty = group.add_subgroups_to_group(added_subgroups)

			notifications.append((("subgroups_changed", group),
			                      dict(added=added_subgroups, removed=removed_subgroups)))

		if default is not None:
			group.change_default(default)
			self._dirty = True

		if self._dirty:
			if save:
				self._save()

			if notify:
				for args, kwargs in notifications:
					self._notify_listeners(*args, **kwargs)


class GroupAlreadyExists(Exception):
	def __init__(self, key):
		Exception.__init__(self, "Group %s already exists" % key)


class UnknownGroup(Exception):
	def __init__(self, key):
		Exception.__init__(self, "Unknown group: %s" % key)


class GroupUnremovable(Exception):
	def __init__(self, key):
		Exception.__init__(self, "Group can't be removed: %s" % key)


class GroupCantBeChanged(Exception):
	def __init__(self, key):
		Exception.__init__(self, "Group can't be changed: %s" % key)


class Group(object):
	def __init__(self, key, name, description="", permissions=None, subgroups=None, default=False, removable=True, changeable=True, toggleable=True):
		self._key = key
		self._name = name
		self._description = description
		self._permissions = permissions
		self._subgroups = subgroups
		self._default = default
		self._removable = removable
		self._changeable = changeable
		self._toggleable = toggleable

	def as_dict(self):
		from octoprint.access.permissions import OctoPrintPermission
		return dict(
			key=self.key,
			name=self.get_name(),
			description=self._description,
			permissions=list(map(lambda p: p.key, self._permissions)),
			subgroups=list(map(lambda g: g.key, self._subgroups)),
			needs=OctoPrintPermission.convert_needs_to_dict(self.needs),
			default=self._default,
			removable=self._removable,
			changeable=self._changeable,
			toggleable=self._toggleable
		)

	@property
	def key(self):
		return self._key

	def get_name(self):
		return self._name

	def get_description(self):
		return self._description

	def is_default(self):
		return self._default

	def is_changeable(self):
		return self._changeable

	def is_removable(self):
		return self._removable

	def is_toggleable(self):
		return self._toggleable

	def add_permissions_to_group(self, permissions):
		"""Adds a list of permissions to a group"""
		if not self.is_changeable():
			raise GroupCantBeChanged(self.key)

		# Make sure the permissions variable is of type list
		if not isinstance(permissions, list):
			permissions = [permissions]

		assert(all(map(lambda p: isinstance(p, OctoPrintPermission), permissions)))

		dirty = False
		for permission in permissions:
			if permissions not in self.permissions:
				self._permissions.append(permission)
				dirty = True

		return dirty

	def remove_permissions_from_group(self, permissions):
		"""Removes a list of permissions from a group"""
		if not self.is_changeable():
			raise GroupCantBeChanged(self.key)

		# Make sure the permissions variable is of type list
		if not isinstance(permissions, list):
			permissions = [permissions]

		assert(all(map(lambda p: isinstance(p, OctoPrintPermission), permissions)))

		dirty = False
		for permission in permissions:
			if permission in self._permissions:
				self._permissions.remove(permission)
				dirty = True

		return dirty

	def add_subgroups_to_group(self, subgroups):
		"""Adds a list of subgroups to a group"""
		if not self.is_changeable():
			raise GroupCantBeChanged(self.key)

		# Make sure the subgroups variable is of type list
		if not isinstance(subgroups, list):
			subgroups = [subgroups]

		assert(all(map(lambda g: isinstance(g, Group), subgroups)))

		dirty = False
		for group in subgroups:
			if group.is_toggleable() and group not in self._subgroups:
				self._subgroups.append(group)
				dirty = True

		return dirty

	def remove_subgroups_from_group(self, subgroups):
		"""Removes a list of subgroups from a group"""
		"""Adds a list of subgroups to a group"""
		if not self.is_changeable():
			raise GroupCantBeChanged(self.key)

		# Make sure the subgroups variable is of type list
		if not isinstance(subgroups, list):
			subgroups = [subgroups]

		assert (all(map(lambda g: isinstance(g, Group), subgroups)))

		dirty = False
		for group in subgroups:
			if group.is_toggleable() and group in self._subgroups:
				self._subgroups.remove(group)
				dirty = True

		return dirty

	def change_default(self, default):
		"""Changes the default flag of a Group"""
		if not self.is_changeable():
			raise GroupCantBeChanged(self.key)

		self._default = default

	def change_description(self, description):
		"""Changes the description of a group"""
		self._description = description

	@property
	def permissions(self):
		if Permissions.ADMIN in self._permissions:
			return Permissions.all()

		return list(filter(lambda p: p is not None, self._permissions))

	@property
	def subgroups(self):
		return filter(lambda g: g is not None, self._subgroups)

	@property
	def needs(self):
		needs = set()
		needs.add(GroupNeed(self.key))
		for p in self.permissions:
			needs = needs.union(p.needs)
		for g in self.subgroups:
			needs = needs.union(g.needs)

		return needs

	def has_permission(self, permission):
		if Permissions.ADMIN.get_name() in self._permissions:
			return True

		return permission.needs.issubset(self.needs)

	def __repr__(self):
		return '{}("{}", "{}", description="{}", permissions={!r}, ' \
		       'default={}, removable={}, changeable={})'.format(self.__class__.__name__,
                                                                 self._key,
		                                                         self._name,
		                                                         self._description,
		                                                         self._permissions,
		                                                         bool(self._default),
		                                                         bool(self._removable),
		                                                         bool(self._changeable))

	def __hash__(self):
		return self.key.__hash__()

	def __eq__(self, other):
		return isinstance(other, Group) and other.key == self.key
